#
# Authors: MN
#
import numpy as np
import spgrep
import spglib
from yambopy.dbs.latticedb import YamboLatticeDB
from yambopy.tools.citations import citation

class Crystal_Symmetries:
    @citation("A. Togo et al., Spglib, Sci. Technol. Adv. Mater. Meth. 4 (2024)")
    def __init__(self, latdb, magnetic_moments=None, tol=1e-5):
        """
        Class to handle  crystal symmetries
        #
        Given a lattice database, finds all symmetries of a  crystal.
        Works irrespective of whether the SAVE has symmetries or not.
        #
        Parameters:
        -----------
        latdb : object
            Lattice database object with required attributes:
        magnetic_moments : array (natom) or (natom,3) or None, optional
            magnetic_moments of each atom (default: None)
            In collinear case, natom is sufficient, but in non-collinear 
            case provide (mx, my, mz) for each atom.
        tol : float, optional
            Symmetry tolerance (default: 1e-5)
        Members:
        --------
        dataset : dict
            The full symmetry dataset from spglib
        spacegroup_type : dict
            Space group type information
        rotations : ndarray
            Rotation matrices in Cartesian coordinates (n_sym, 3, 3)
        translations : ndarray
            Translation vectors in Cartesian coordinates (n_sym, 3)
        international_symbol : str
            International symbol of the space group
        hall_symbol : str
            Hall symbol of the space group
        wyckoffs : list
            Wyckoff letters for each atom
        pointgroup : str
            Point group symbol
        spacegroup_number : int
            International space group number
        """
        # Get lattice and atomic information
        lattice = latdb.lat
        numbers = np.rint(latdb.atomic_numbers).astype(int)
        lat_inv = np.linalg.inv(lattice)
        positions = latdb.car_atomic_positions @ lat_inv
        positions = positions - np.floor(positions)
        #
        self.magnetic = not (magnetic_moments is None)
        #
        # Get symmetry dataset
        if self.magnetic :
            cell = (lattice, positions, numbers, magnetic_moments)
            print("Currently Magnetic systems not supported.")
            exit()
        else :
            cell = (lattice, positions, numbers)
            self.dataset = spglib.get_symmetry_dataset(cell, symprec=tol)
            self.spacegroup_type = spglib.get_spacegroup_type(self.dataset.hall_number)
        #
        # Convert rotation and fractional translations to Cartesian units
        self.rotations = lattice.T[None,:,:] @ self.dataset.rotations @ lat_inv.T[None,:,:]
        self.translations = self.dataset.translations @ lattice
        #
        self.international_symbol = self.dataset.international
        self.hall_symbol = self.dataset.hall
        self.wyckoffs = self.dataset.wyckoffs
        self.pointgroup = self.dataset.pointgroup
        self.pointgroup_schoenflies = self.spacegroup_type.pointgroup_schoenflies
        self.spacegroup_number = self.dataset.number
    #
    #
    def write_symmetry_info_to_file(self, filename="sym_ops.dat"):
        """Write symmetries in cart units to file."""
        sym_matrices = self.rotations
        trans_vectors = self.translations
        # 
        with open(filename, "w") as f:
            # Write basic info
            f.write("# Basic symmetry information generated by Yambopy using spglib.\n")
            f.write("# International symbol : %s\n" %(str(self.international_symbol)))
            f.write("# Hall symbol          : %s\n" %(str(self.hall_symbol)))
            f.write("# Wyckoffs             : %s\n" %(str(self.wyckoffs)))
            f.write("# Pointgroup           : %s (%s)\n" %(str(self.pointgroup),str(self.pointgroup_schoenflies)))
            f.write("# Spacegroup           : %s\n" %(str(self.spacegroup_number)))
            #
            f.write("# \n")
            f.write("# Coset representatives \n")
            f.write("# \n")
            #
            f.write("# index         symmetry matrix (cart units).         fractional translation (cart units)\n")
            f.write("# -------------------------------------------------------------------------------------\n")

            for i, (S, v) in enumerate(zip(sym_matrices, trans_vectors), start=1):

                # First matrix row printed on same line as index
                row0 = " ".join(f"{S[0, c]:10.6f}" for c in range(3))
                line = f"{i:>{4}d}" + " " * (12 - 4) + row0
                f.write(line + "\n")

                # Second row includes translation vector
                row1 = " ".join(f"{S[1, c]:10.6f}" for c in range(3))
                vec  = " ".join(f"{v[k]:10.6f}" for k in range(3))
                line = " " * 12 + row1
                line += " " * (52 - len(line)) + vec
                f.write(line + "\n")

                # Third row, matrix only
                row2 = " ".join(f"{S[2, c]:10.6f}" for c in range(3))
                line = " " * 12 + row2
                f.write(line + "\n")

                f.write("#-------------------------------------------------------------------------------------\n")


## Test
if __name__ == "__main__":
    import os
    np.set_printoptions(suppress=True)
    latdb = YamboLatticeDB.from_db_file(os.path.join('.', 'SAVE'))
    symm = Crystal_Symmetries(latdb,tol=1e-4)
    print(symm.wyckoffs)
    print(symm.pointgroup_schoenflies)




from yambopy import *
import sys
import argparse

if __name__ == "__main__":
    #parse options
    parser = argparse.ArgumentParser(description='RT setup')
    parser.add_argument('-f', '--field_direction',nargs='+',type=float,help='<Required> Set field direction',required=True)    
    parser.add_argument('-p', '--prefix',type=str,help='<Required> QE prefix',required=True)

    args = parser.parse_args()

    if len(sys.argv)<=2:
        parser.print_help()
        sys.exit(1)

    CreateYamboSave(args.prefix,field_dir=args.field_direction)



#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: HPC, AMS, FP, RR
#
# This file is part of the yambopy project
#
import numpy as np

def CmplxType(var):
    """ Distinguish between double and float for storing residuals and eigenvector with the same precision as in the Yambo database
    """
    if var.dtype=='float32':   return np.complex64
    elif var.dtype=='float64': return np.complex128
    else: raise TypeError('\n[ERROR] Variable type not recognized. It should be either float (float32) or double (float64).\n')



#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: FP, HPC
#
# This file is part of the yambopy project
#
import numpy as np
from itertools import product
from yambopy.lattice import red_car, vec_in_list, isbetween, car_red
from qepy.lattice import Path
from scipy.spatial import KDTree

def expand_kpoints(car_kpoints,sym_car,rlat,atol=1.e-6):
    """
    Take a list of kpoints and symmetry operations and return the full brillouin zone
    with the corresponding index in the irreducible brillouin zone

    Input:
    * IBZ kpoints in Cartesian coodinates
    * Symmetry operations in Cartesian coordinates
    * rlat: reciprocal lattice vectors
    * atol: tolerance for the distance between kpoints

    Output:
    * weights: weights of the kpoints in the irreducible brillouin zone
    * kpoints_indexes: indexes of the kpoints in the irreducible brillouin zone
    * symmetry_indexes: indexes of the symmetries used to generate the kpoints
    * kpoints_full: kpoints in the full brillouin zone
    """

    #check if the kpoints were already exapnded
    kpoints_indexes  = []
    kpoints_full     = []
    symmetry_indexes = []

    #kpoints in the full brillouin zone organized per index
    kpoints_full_i = {}

    #expand using symmetries
    for nk,k in enumerate(car_kpoints):
        #if the index in not in the dicitonary add a list
        if nk not in kpoints_full_i:
            kpoints_full_i[nk] = []

        for ns,sym in enumerate(sym_car):

            new_k = np.dot(sym,k)
            #check if the point is inside the bounds
            k_red = car_red([new_k],rlat)[0]
            k_red[np.abs(k_red) < atol] = 0. # Set to zero values < atol to avoid mistakes
            k_bz = (k_red+atol)%1

            #if the vector is not in the list of this index add it
            if not vec_in_list(k_bz,kpoints_full_i[nk],atol=atol):
                kpoints_full_i[nk].append(k_bz)
                kpoints_full.append(new_k)
                kpoints_indexes.append(nk)
                symmetry_indexes.append(ns)
                continue

    #calculate the weights of each of the kpoints in the irreducible brillouin zone
    nkpoints_full = len(kpoints_full)
    weights = np.zeros([nkpoints_full])
    for nk in kpoints_full_i:
        weights[nk] = float(len(kpoints_full_i[nk]))/nkpoints_full

    return np.array(weights), np.array(kpoints_indexes), np.array(symmetry_indexes), np.array(kpoints_full)

def get_path_car(kpts_path_car,path):
    """
    Convert a Path object in reduced coordinates to a path in Cartesian coordinates
    """
    return Path( [[kpts_path_car[i],path.klabels[i]] for i in range(len(kpts_path_car))],path.intervals )

def get_path(car_kpoints,rlat,sym_car,path,debug=False):
    """
    Obtain a list kpoints along a specific high-symmetry path

    Input:
    * car_kpoints: kpoints in Cartesian coordinates [IBZ/Full BZ]
    * rlat: reciprocal lattice vectors
    * sym_car [symmetry ops. if car_kpoints given in IBZ] | None [if car_kpoints in fulll BZ]
    * path: Path object with the high-symmetry path (points in reduced coordinates)

    Output:
    * bands_kpoints: kpoints Cartesian coordinates along the path
    * bands_indexes: indexes of the kpoints in the path
    * path_car: path in Cartesian coordinates
    """

    # expand if symmetries are provided, otherwise the kpoints are considered already expanded
    if sym_car is None: nks = list(range(len(car_kpoints)))
    else:               _, nks, _, car_kpoints = expand_kpoints(car_kpoints,sym_car,rlat)

    # high-symmetry points in cartesian coordinates
    kpts_path_car = red_car(path.kpoints, rlat)
    # Path object in cartesian coordinates (for later plotting)
    path_car = get_path_car(kpts_path_car,path)

    #find the points along the high symmetry lines
    bands_kpoints = []
    bands_indexes = []

    #for all the paths
    for k in range(len(path.kpoints)-1):

        # store here all the points in the path
        # key:   has the coordinates of the kpoint rounded to 4 decimal places
        # value: index of the kpoint
        #        distance to the starting kpoint
        #        the kpoint cordinate
        kpoints_in_path = {}

        start_kpt = kpts_path_car[k]   #start point of the path
        end_kpt   = kpts_path_car[k+1] #end point of the path

        #generate repetitions of the brillouin zone
        for x,y,z in product(list(range(-1,2)),list(range(-1,2)),list(range(1))):

            #shift the brillouin zone
            shift = red_car([np.array([x,y,z])],rlat)[0]

            #iterate over all the kpoints
            for index, kpt in zip(nks,car_kpoints):

                kpt_shift = kpt+shift #shift the kpoint

                #if the point is collinear we add it
                if isbetween(start_kpt,end_kpt,kpt_shift):
                    key = tuple([round(kpt,4) for kpt in kpt_shift])
                    value = [ index, np.linalg.norm(start_kpt-kpt_shift), kpt_shift ]
                    kpoints_in_path[key] = value

        #sort the points acoording to distance to the start of the path
        kpoints_in_path = sorted(list(kpoints_in_path.values()),key=lambda i: i[1])

        #for all the kpoints in the path
        for index, disp, kpt in kpoints_in_path:
            bands_kpoints.append( kpt )
            bands_indexes.append( index )
            if debug: print(("%12.8lf "*3)%tuple(kpt), index)

    return bands_kpoints, bands_indexes, path_car


def make_kpositive(klist, tol=1e-6):
    """
    Shifts all k-points into the range [0,1) by applying modulo operation.

    Parameters
    ----------
    klist : np.ndarray
        Array of k-points in crystal coordinates.
    tol : float, optional
        A small tolerance to ensure numerical stability, default is 1e-6.

    Returns
    -------
    np.ndarray
        Array of k-points mapped to the range [0,1).
    """
    kpos = klist - np.floor(klist)  # Ensure k-points are within [0,1)
    return (kpos + tol) % 1  # Apply small tolerance correction


def build_ktree(kpts):
    """
    Builds a k-d tree for efficient k-point searching.

    Parameters
    ----------
    kpts : np.ndarray
        Array of k-points in crystal coordinates.

    Returns
    -------
    KDTree
        A KDTree structure for fast nearest-neighbor lookup of k-points.
    """
    tree = make_kpositive(kpts)  # Normalize k-points to [0,1)
    return KDTree(tree, boxsize=[1, 1, 1])  # Construct KDTree with periodic boundaries


def find_kpt(tree, kpt_search, tol=1e-5):
    """
    Finds the index of a k-point in a KDTree within a given tolerance.

    Parameters
    ----------
    tree : KDTree
        Pre-built KDTree of k-points.
    kpt_search : np.ndarray
        The k-point to search for in crystal coordinates.
    tol : float, optional
        Tolerance for k-point matching, default is 1e-5.

    Returns
    -------
    int
        Index of the matched k-point in the original k-point list.

    Raises
    ------
    SystemExit
        If the k-point is not found within the specified tolerance.
    """
    kpt_search = make_kpositive(kpt_search)  # Normalize k-point
    dist, idx = tree.query(kpt_search, workers=1)  # Perform nearest-neighbor search
    assert np.max(dist) < tol, "Kpoint not found"
    return idx  # Return the index of the found k-point

def regular_grid(nk1,nk2,nk3):
    """
    Generation of positive-coordinate full 'square' grid starting from zero
    Shifted grids not allowed for now
    """
    i, j, k = np.meshgrid(np.arange(nk1), np.arange(nk2), np.arange(nk3), indexing='ij')
    xkg = np.array([
        i.flatten() / nk1,
        j.flatten() / nk2,
        k.flatten() / nk3,
    ])
    return xkg.T # shape [nk,3]

def find_kpatch(kpts, kcentre, kdist, lat_vecs):
    """
    find set of kpoints around the kcentre with in kdist

    Parameters
    ----------
    kpts : kpoints in crystal coordinates (nk,3)
    kcentre : kpoint centre in crystal coordinates (3)
    kdist : distance around kcentre to be considered in atomic units 
            i.e 1/bohr.
    lat_vecs: lattice vectors. ith lattice vector is ai = a[:,i]
    Returns
    -------
    int array
        Indices of kpoints in kpts array which satify the given condition i.e
        | k - kcentre + G0| <= kdist, where G0 is reciprocal lattice vector to bring to BZ
    """
    #
    blat = 2*np.pi*np.linalg.inv(lat_vecs)
    kdiff = kpts-kcentre[None,:]
    kdiff = kdiff-np.floor(kdiff)
    #
    tmp_arr = np.array([-3, -2, -1, 0, 1, 2, 3])
    nG0 = len(tmp_arr)
    G0 = np.zeros((nG0,nG0,nG0,3))
    G0[...,0], G0[...,1], G0[...,2] = np.meshgrid(tmp_arr, tmp_arr,
                                                  tmp_arr, indexing='ij')
    G0 = G0.reshape(-1,3)
    kdiff = kdiff[:,None,:]-G0[None,:,:]
    kdiff = kdiff.reshape(-1,3)@blat
    kdiff = np.linalg.norm(kdiff,axis=-1).reshape(len(kpts),-1)
    kdiff = np.min(kdiff,axis=-1)
    return np.where(kdiff <= kdist)[0]

    
def generate_kpoint_grid(nk1,nk2,nk3,sym_and_trev,IBZ=True,eps=1.0e-5):
    """
    Generation of gamma-centered Monkhorst-Pack grid.

    This function is the python porting of the Quantum ESPRESSO subroutine `kpoint_grid` found in PW/kpoint_grid.f90

    Input:
        nk1, nk2, nk3 -> grid dimensions
        sym_and_trev  -> (sym_red, time_rev, time_rev_list)

        (Symmetries in reduced coordinates, trev logical, indices of symmetries composed with time reversal)
        These can be obtained from the attributes of the same name in YamboLatticeDB

        IBZ [True]    -> grid is reduced to IBZ by checking equivalent points
            [False]   -> grid is generated with symmetries turned off
                         (i.e. NOT in BZ - the Wigner-Seitz cell - but in the primitive reciprocal unit cell)
        epes [1e-5]   -> numerical error for equivalent points

        In order to obtain a list of kpoints in the full BZ, use IBZ==True and then expand_kpoints().

    Output:
        xks -> Number of kpoints in IBZ
        xk  -> kpoints in IBZ/BZ in CRYSTAL coordinates (convert to cc with red_car())
        wk  -> kpoints weights
    """
    Nk    = nk1*nk2*nk3
    wkk   = np.ones(Nk)    # Weights
    equiv = np.arange(Nk)  # Initially, each kpoint is equivalent to itself

    # symmetry and time-reversal info
    sym_red, is_trev, is_sym_trev = sym_and_trev
    Nsym = len(sym_red)

    # Generate full regular grid in crystal coordinates
    xkg = regular_grid(nk1,nk2,nk3) # [nk,3]

    if IBZ:
        # Now we have to start checking for equivalent points:
        for ik in range(Nk):
            # Check if this k-point has already been found equivalent to another
            # Do not consider points previously found equivalent to another
            # Check both k and -k
            if equiv[ik] == ik:
                for i_s in range(Nsym):
                    # Apply symmetry operations
                    xkr = np.dot(sym_red[i_s,:,:], xkg[ik,:])
                    # Bring back in 1st BZ
                    xkr -= np.round(xkr)
                    # Take opposite if symmetry is composed with TR
                    if is_sym_trev[i_s]: xkr = -xkr

                    # Check if in the list
                    xx, yy, zz = xkr*[nk1,nk2,nk3]
                    in_the_list = all( abs(v-round(v)) <= eps for v in [xx, yy, zz])

                    if in_the_list:
                        i,j,k = [(round(xkr[dim]*nki + 2*nki) % nki) + 1 for dim, nki in enumerate([nk1, nk2, nk3])]
                        n = (k-1) + (j-1)*nk3 + (i-1)*nk2*nk3

                        if n > ik and equiv[n] == n:
                            equiv[n] = ik
                            wkk[ik] += 1.0
                        elif equiv[n] != ik or n < ik:
                            raise ValueError("Error in the checking algorithm")

                    # Time reversal symmetry check
                    if is_trev:
                        xx, yy, zz = -xkr*[nk1,nk2,nk3]
                        in_the_list = all(abs(v - round(v)) <= eps for v in [xx, yy, zz])

                        if in_the_list:
                            i, j, k = [(round(-xkr[dim]*nki + 2*nki) % nki) + 1 for dim, nki in enumerate([nk1, nk2, nk3])]
                            n = (k-1) + (j-1)*nk3 + (i-1)*nk2*nk3

                            if n > ik and equiv[n] == n:
                                equiv[n] = ik
                                wkk[ik] += 1.0
                            elif equiv[n] != ik or n < ik:
                                raise ValueError("Error in the checking algorithm")

    # Filter unique k-points
    unique_kpoints = (equiv == np.arange(Nk))
    # Bring back unique points into first BZ
    xk = xkg[unique_kpoints,:] - np.round(xkg[unique_kpoints,:])
    wk = wkk[unique_kpoints] / np.sum(wkk[unique_kpoints])  # Normalize weights
    nks = len(xk[0])

    return nks, xk, wk



#
# Authors: MN
#
import os
import numpy as np
from yambopy.dbs.excitondb import YamboExcitonDB
from yambopy.dbs.latticedb import YamboLatticeDB
from yambopy.dbs.wfdb import YamboWFDB
from yambopy.bse.rotate_excitonwf import rotate_exc_wf
from yambopy.tools.degeneracy_finder import find_degeneracy_evs
from yambopy.symmetries.point_group_ops import get_pg_info, decompose_rep2irrep
from yambopy.symmetries.crystal_symmetries import Crystal_Symmetries
from yambopy.tools.citations import citation

@citation("M. Nalabothula et al. arXiv:2511.21540 (2025)")
def compute_exc_rep(path='.', bse_dir='SAVE', iqpt=1, nstates=-1, degen_tol = 1e-3,
                    degen_rtol=1e-3, symm_tol=1e-2, use_save_symmetries=False):
    """
    Perform a group–theoretical analysis of excitonic wavefunctions
    at a given BSE q-point using the crystal symmetries.

    The routine:
      * Loads lattice, exciton, and electron wave-function databases
      * Identifies positive-energy excitons and degeneracy groups
      * Determines the little group of the chosen q-point using spglib
      * Rotates the exciton wavefunctions under each symmetry
      * Builds the representation matrices of the little group
      * Decomposes each excitonic multiplet into irreducible
        representations of the corresponding point group

    .. note::
       For finite q ≠ 0, use ``Lkind="full"`` (default in Yambo).
       For q = 0, ``Lkind="bar"`` should be used as ``"full"``
       breaks symmetries depending on the gauge.

       For magnetic non-collinear systems, this function falls back to SAVE symmetries.
       This is because, spglib requires magnetic moment  for each atom and there
       is no way we can get it from SAVE as of now. Ofcourse, the price we pay is that
       non-symmorphic symmetries are not taken into account and there by much smaller groups.


    Parameters
    ----------
    path : str, optional
        Working directory containing the ``SAVE`` folder.
        Default is current directory ``'.'``.
    bse_dir : str, optional
        Directory containing the BSE diagonalization database
        (e.g. ``SAVE`` or ``GW_BSE``). Default ``'SAVE'``.
    iqpt : int, optional
        Index of the q-point in the BSE calculation (1-based as in Yambo).
        Default: ``1``.
    nstates : int, optional
        Number of positive-energy exciton states to analyze.
        If ``-1`` (default) all positive states are used.
    degen_tol : float, optional
        Absolute tolerance to detect degeneracies in the exciton energies.
        Default ``1e-3``.
    degen_rtol : float, optional
        Relative tolerance to detect degeneracies in the exciton energies.
        Default ``1e-3``.
    symm_tol : float, optional
        Tolerance to Symmetry operators.
        Default ``1e-2``.
    use_save_symmetries : bool, optional
        If True, uses internal symmetries from SAVE instead of spglib symmetries.

    Outputs
    -------
    No return value.
    Prints to stdout:
        * Identified little group and symmetry operations
        * Degenerate groups of excitons (energies and multiplicities)
        * Irreducible-representation decomposition of each multiplet

    Files Cached in BSE directory (if not present)
    -----------------------------
    ``BS_left_ev_Cache.npy`` :
        Left eigenvectors of the BSE Hamiltonian in case of non-TDA.
    ``Dmat_elec_Cache_spglib.npy`` :
        Electronic representation matrices from spglib symmetry operations.

    Note :
    Currently projective representations are not implemented and left for future.

    Raises
    ------
    AssertionError
        1) If no positive-energy exciton eigenvalues are found.
        2) Projective representations for non-symmorphic symmetires at boundary Q points

    Examples
    --------
    >>> compute_exc_rep(path='..', bse_dir='../GW_BSE',
    ...                  iqpt=1, nstates=3, degen_tol=1e-2)

    """
    if nstates == 0 :
        print("Warning : Number of states set to 0. so returning without any output.")
        return None
    # Load the lattice database
    lattice = YamboLatticeDB.from_db_file(os.path.join(path, 'SAVE', 'ns.db1'))
    # NM  : For now fall back to save symmetries as we donot have information on magnetic moments of atoms
    # to get the magnetic symmetries of the crystal.
    if lattice.mag_syms and not use_save_symmetries:
        print("Warning : For magnetic symmetries, the code falls back to symmetries in the SAVE")
        use_save_symmetries = True
    #
    ## Get spglib symmetries to have full symmtries of the crystal even though they are not in yambo base
    # Write basic symmetry information to file
    Rotation_matrices_symm = lattice.sym_car
    frac_trans_symm = np.zeros((len(Rotation_matrices_symm),3),dtype=Rotation_matrices_symm.dtype)
    time_rev = int(np.rint(lattice.time_rev))
    if not use_save_symmetries:
        symm = Crystal_Symmetries(lattice,tol=1e-4)
        symm_info_file_name = "Symmetry_info_excitons.txt"
        print("Writing Basic symmetry information to file : ",symm_info_file_name)
        symm.write_symmetry_info_to_file(symm_info_file_name)
        Rotation_matrices_symm = symm.rotations
        frac_trans_symm = symm.translations
        time_rev = False
    #
    # NM : I am assuming that we can have atmost 100 accidental degeneracies (only for TDA).
    # This is a fine if we are not in continum. In that case, use should increase more states.
    nstates_read = -1
    if nstates >0: nstates_read = nstates + 100
    #
    filename = 'ndb.BS_diago_Q%d' % (iqpt)
    excdb = YamboExcitonDB.from_db_file(lattice, filename=filename,
                                             folder=bse_dir,
                                             Load_WF=True, neigs=nstates_read)
    # Check for Lkind
    ## Lkind = "full" for q !=0, as Lkind = "bar" should not be used for finite q
    ## for q = 0, Lkind = "bar" is recommended as "full" will 
    ## break symmetries (depends on the chosen direction)
    if excdb.Lkind is not None: # compatibility with old databases
        if iqpt==0 and 'bar' not in excdb.Lkind:
            print(f"Warning: You are using Lkind={excdb.Lkind} at q=0, but 'Lbar' is recommended. Symmetries are broken in this way.")
        if iqpt!=0 and 'full' not in excdb.Lkind:
            print(f"Warning: You are using Lkind={excdb.Lkind} at q!=0, but 'Lfull' should be used.")

    # Load the wavefunction database
    wfdb = YamboWFDB(path=path, latdb=lattice,
                      bands_range=[np.min(excdb.table[:, 1]) - 1,
                    np.max(excdb.table[:, 2])])
    Akcv = excdb.get_Akcv()
    Akcv_left = Akcv
    if Akcv.shape[1] == 2:
        left_ev_path = os.path.join(bse_dir,f'BS_left_ev_Cache_qpt_{iqpt}.npy')
        if os.path.exists(left_ev_path):
            print("Found left eigenvectors, loading ...")
            Akcv_left = np.load(left_ev_path)
        else:
            print("Computing left ev ...")
            Akcv_left = np.linalg.inv(Akcv.reshape(len(Akcv),-1)).conj().T.reshape(Akcv.shape)
            np.save(left_ev_path,Akcv_left)
    #
    eigs = excdb.eigenvalues.real
    sort_idx = np.argsort(eigs)
    eigs = eigs[sort_idx].copy()
    pos_idx = np.where(eigs > 0)
    assert len(pos_idx) >0, "No postive eigenvalues found"
    pos_idx = pos_idx[0][0]
    #
    n_pos = len(eigs[pos_idx:])
    #
    if nstates < 0: nstates = n_pos
    else : nstates = min(nstates, n_pos)
    #
    if nstates < n_pos:
        last = eigs[pos_idx + nstates - 1]
        tol = degen_tol + degen_rtol * abs(last)
        extra = np.sum(np.abs(eigs[pos_idx+nstates:pos_idx+n_pos] - last) < tol)
        if extra: print("Warning: More excitonic states included due to degeneracies.")
        nstates += extra
    #
    Ak_r = Akcv[sort_idx][pos_idx:pos_idx+nstates]
    Ak_l = Akcv_left[sort_idx][pos_idx:pos_idx+nstates].conj()
    exe_ene = eigs[pos_idx:pos_idx+nstates]
    #
    degen_idx = find_degeneracy_evs(exe_ene,atol=degen_tol,rtol=degen_rtol)
    uni_eigs = []
    degen_eigs = []
    for i in degen_idx:
        uni_eigs.append(np.mean(exe_ene[i]))
        degen_eigs.append(len(i))
    uni_eigs = np.array(uni_eigs)
    degen_eigs = np.array(degen_eigs,dtype=int)

    excQpt = excdb.car_qpoint
    # Convert the q-point to crystal coordinates
    excQpt = lattice.lat @ excQpt

    lat_vec = lattice.lat
    lat_vec_inv = np.linalg.inv(lat_vec)
    #
    trev_fac = 1 + int(time_rev)
    # The number of symmetries for which we need representations matrices.
    # We only need representations for spatial symmetries.
    # Note yambo always orders time rev symmeties in second half, even for magnetic systems.
    nsym_spatial = len(Rotation_matrices_symm)//trev_fac
    #
    Dmat_path = os.path.join(bse_dir, 'Dmat_elec_Cache_spglib.npy')
    if use_save_symmetries: Dmat_path = os.path.join(bse_dir, 'Dmat_elec_Cache_SAVE.npy')
    #
    if os.path.exists(Dmat_path):
        print("Dmats found. Loading ....")
        dmats = np.load(Dmat_path)
    else:
        print("Dmats not found. Computing ....")
        dmats = wfdb.Dmat(symm_mat=Rotation_matrices_symm[:nsym_spatial],
                          frac_vec=frac_trans_symm[:nsym_spatial], time_rev=False)
        np.save(Dmat_path,dmats)
    #
    ## print some data about the degeneracies
    print('=' * 40)
    print('Group theory analysis for Q point : (%.6f, %.6f, %.6f)' %(excQpt[0],excQpt[1],excQpt[2]))
    print('*' * 40)

    trace_all_real = []
    trace_all_imag = []
    little_group = []
    #
    #
    for isym in range(nsym_spatial):
        symm_mat = Rotation_matrices_symm[isym]
        symm_mat_red = lat_vec@symm_mat@lat_vec_inv
        #
        Sq_minus_q = np.einsum('ij,j->i', symm_mat_red, excQpt) - excQpt
        #print(Sq_minus_q)
        #diff = Sq_minus_q.copy()
        Sq_minus_q = Sq_minus_q - np.rint(Sq_minus_q)
        ## check if Sq = q
        if np.linalg.norm(Sq_minus_q) > 10**-5: continue
        little_group.append(isym + 1)
        tau_dot_k = np.exp(1j * 2 * np.pi *
                       np.dot(excdb.car_qpoint, frac_trans_symm[isym]))
        #assert(np.linalg.norm(Sq_minus_q)<10**-5)
        rot_Akcv = rotate_exc_wf(Ak_r, symm_mat_red, wfdb.kBZ, excQpt,
                                 dmats[isym], False, ktree=wfdb.ktree)
        rep = tau_dot_k*np.einsum('m...,n...->mn',Ak_l,rot_Akcv,optimize=True)
        # NM : In case of non-trivial projective irrep, we cannot do this,
        # so this fails for boundary Q points, for example (0,0,0.5) in bulk hBN.
        ## Check if this is projective rep and exit as it is yet not implemented.
        G0_tmp = np.einsum('ji,j->i', symm_mat, excdb.car_qpoint) - excdb.car_qpoint
        G0_tau_tmp = G0_tmp.dot(frac_trans_symm[isym])
        G0_tau_tmp = G0_tau_tmp-np.rint(G0_tau_tmp)
        if np.abs(G0_tau_tmp) > 1e-3:
            exit("Error : Projective representations are not implemented")
        #print('Symmetry number : ',isym + 1)
        ## print characters
        irrep_sum = 0
        real_trace = []
        imag_trace = []
        for iirepp in range(len(uni_eigs)):
            idegen = degen_eigs[iirepp]
            idegen2 = irrep_sum + idegen
            trace_tmp = np.trace(rep[irrep_sum:idegen2, irrep_sum:idegen2])
            real_trace.append(trace_tmp.real.round(4))
            imag_trace.append(trace_tmp.imag.round(4))
            irrep_sum = idegen2
        # print('Real : ',real_trace)
        # print('Imag : ',imag_trace)
        trace_all_real.append(real_trace)
        trace_all_imag.append(imag_trace)

    little_group = np.array(little_group, dtype=int)
    #
    pg_label, classes, class_dict, char_tab, irreps = get_pg_info(
        Rotation_matrices_symm[little_group - 1])

    print('Little group : ', pg_label)
    print('Little group symmetries : ', little_group)
    # print class info
    print('Classes (symmetry indices in each class): ')
    req_sym_characters = np.zeros(len(classes), dtype=int)
    class_orders = np.zeros(len(classes), dtype=int)
    for ilab, iclass in class_dict.items():
        print("%16s    : " % (classes[ilab]), little_group[np.array(iclass)])
        req_sym_characters[ilab] = min(iclass)
        class_orders[ilab] = len(iclass)
    print()
    trace_all_real = np.array(trace_all_real)
    trace_all_imag = np.array(trace_all_imag)
    trace = trace_all_real + 1j * trace_all_imag
    trace_req = trace[req_sym_characters, :].T
    print("====== Exciton representations ======")
    print("Energy (eV),  degeneracy  : representation")
    print('-' * 40)
    for i in range(len(trace_req)):
        rep_str_tmp = decompose_rep2irrep(trace_req[i], char_tab, len(little_group),
                                          class_orders, irreps,tol=symm_tol)
        print('%.4f        %9d  : ' % (uni_eigs[i], degen_eigs[i]), rep_str_tmp)
    print('*' * 40)



if __name__ == "__main__":
    compute_exc_rep(path='..', bse_dir='GW_BSE', iqpt=1, nstates=3, degen_tol = 1e-2)




#
# This file is part of yambopy
#
import numpy as np
from qepy import xml

"""
This file contains a list of functions to deal with Bravais lattices
"""

def get_ibrav(datafile_xml):
    """
    Get ibrav number (as string) from QE data-file-schema.xml
    
    [NB] datafile_xml must be in the form:
    
        datafile_xml = ET.parse( "PREFIX.save/data-file-schema.xml" )
    """
    
    ibrav = xml.get_xml_attrib(datafile_xml,'atomic_structure','bravais_index',repeated=True)
    return ibrav

def lattice_dictionary(ibrav):
    """
    Dictionary with text descriptions for ibrav numbers
    """    
    bravais_lattices = {
    0 : 'free',
    1 : 'cubic P (sc)',
    2 : 'cubic F (fcc)',
    3 : 'cubic I (bcc)',
    4 : 'Hexagonal and Trigonal P',
    5 : 'Trigonal R, 3fold axis c',
    6 : 'Tetragonal P (st)',
    7 : 'Tet ragonal I (bct)',
    8 : 'Orthorhombic P',
    9 : 'Orthorhombic base-centered(bco)',
    10 : 'Orthorhombic face-centered',
    11 : 'Orthorhombic body-centered',
    12 : 'Monoclinic P, unique axis c',
    13 : 'Monoclinic base-centered',
    14 : 'Triclinic'
    }
    try: bravais_lattices[ibrav]
    except KeyError: print("This lattice type is not implemented ):")
    else: return bravais_lattices[ibrav]
    
def lattice_type(ibrav,cell):
    """
    Input: specific lattice type (ibrav) and cell parameters (as list)
    Output: lattice vectors 
    """
    ibrav = int(ibrav)
    lattice = lattice_dictionary(ibrav)
    sqrt = np.sqrt
    
    if lattice=='Hexagonal and Trigonal P':
        a    = cell[0]
        c_a  = cell[2] #c/a
        vecs = a*np.array([ [  1.,          0.,  0. ], 
                            [-0.5, sqrt(3.)/2.,  0. ],
                            [  0.,          0., c_a ] ])
    if lattice=='free':
        vecs = np.array([ datafile_xml.find("CELL/DIRECT_LATTICE_VECTORS/a1").text.split(),
                          datafile_xml.find("CELL/DIRECT_LATTICE_VECTORS/a2").text.split(),
                          datafile_xml.find("CELL/DIRECT_LATTICE_VECTORS/a3").text.split() ]).astype(float)
    if lattice=='cubic P (sc)':
        a    = cell[0]
        vecs = a*np.array([ [ 1., 0., 0. ],
                            [ 0., 1., 0. ],
                            [ 0., 0., 1. ] ])
    if lattice== 'cubic F (fcc)':
        a    = cell[0]
        vecs = a/2.*np.array([ [-1., 0., 1. ],
                               [ 0., 1., 1. ],
                               [-1., 1., 0. ] ]) 
    if lattice=='cubic I (bcc)':
        a    = cell[0]
        vecs = a/2.*np.array([ [ 1., 1., 1. ],
                               [-1., 1., 1. ],
                               [-1.,-1., 0. ] ])
    if lattice=='Trigonal R, 3fold axis c':
        a    = cell[0]
        c_g  = cell[3] #cos(gamma)
        tx, ty, tz = [ sqrt((1.-c_g)/2.), sqrt((1.-c_g)/6.), sqrt((1.+2*c_g)/3.) ]
        vecs = a*np.array([ [ tx,  -ty, tz ],
                            [ 0., 2*ty, tz ],
                            [-tx,  -ty, tz ] ]) 
    if lattice=='Tetragonal P (st)':
        a    = cell[0]
        c_a  = cell[2]
        vecs = a*np.array([ [ 1., 0.,  0. ],
                            [ 0., 1.,  0. ],
                            [ 0., 0., c_a ] ])
    if lattice=='Tetragonal I (bct)':
        a    = cell[0]
        c_a  = cell[2]
        vecs = a/2.*np.array([ [ 1.,-1., c_a ],
                               [ 1., 1., c_a ],
                               [-1.,-1., c_a ] ])
    if lattice=='Orthorhombic P':
        a    = cell[0]
        b_a  = cell[1] #b/a
        c_a  = cell[2]
        vecs = a*np.array([ [ 1.,  0.,  0. ],
                            [ 0., b_a,  0. ],
                            [ 0.,  0., c_a ] ])
    if lattice=='Orthorhombic base-centered(bco)':
        a    = cell[0]
        b_a  = cell[1]
        c_a  = cell[2]
        vecs = a*np.array([ [ 1./2., b_a/2.,  0. ],
                            [-1./2., b_a/2.,  0. ],
                            [    0.,     0., c_a ] ])
    if lattice=='Orthorhombic face-centered':
        a    = cell[0]
        b_a  = cell[1]
        c_a  = cell[2]
        vecs = a/2.*np.array([ [ 1.,  0., c_a ],
                               [ 1., b_a,  0. ],
                               [ 0., b_a, c_a ] ])
    if lattice=='Orthorhombic body-centered':
        a    = cell[0]
        b_a  = cell[1]
        c_a  = cell[2]
        vecs = a/2.*np.array([ [ 1., b_a, c_a ],
                               [-1., b_a, c_a ],
                               [-1.,-b_a, c_a ] ])
    if lattice=='Monoclinic P, unique axis c':
        a    = cell[0]
        b_a  = cell[1]
        c_a  = cell[2]
        c_g  = cell[3]
        s_g  =+sqrt(1.-c_g*c_g)
        vecs = a*np.array([ [      1.,      0.,  0. ],
                            [ b_a*c_g, b_a*s_g,  0. ],
                            [      0.,      0., c_a ] ])
    if lattice=='Monoclinic base-centered':
        a    = cell[0]
        b_a  = cell[1]
        c_a  = cell[2]
        c_g  = cell[3]
        s_g  =+sqrt(1.-c_g*c_g)
        vecs = a*np.array([ [   1./2.,      0.,-c_a/2. ],
                            [ b_a*c_g, b_a*s_g,     0. ],
                            [   1./2.,      0., c_a/2. ] ])
    if lattice=='Triclinic':
        a    = cell[0]
        b_a  = cell[1]
        c_a  = cell[2]
        c_a  = cell[3] #cos(alpha)
        c_b  = cell[4] #cos(beta) 
        c_g  = cell[5] #cos(gamma)
        s_g  =+sqrt(1.-c_g*c_g)
        V2   = a**6.*b_a**2.*c_a**2.*(1.+2.*c_a*c_b*c_g-c_a*c_a-c_b*c_b-c_g*c_g)
        vecs = a*np.array([ [      1.,      0., 0. ],
                            [ b_a*c_g, b_a*s_g, 0. ],
                            [ c_a*c_b, c_a*(c_a-c_b*c_g)/s_g, c_a*sqrt(V2)/(a*b_a*a*c_a*a)/s_g ] ])  
    #
    try: vecs
    except NameError: print("This lattice type is not implemented ):")
    else: return vecs

def crys_to_car(lattice_vectors,reciprocal_space=False):
    """ Calculation of the metric tensor in real or reciprocal space
    
        Input:
            - Lattice vectors (real space)
    """
    
    #Metric tensor
    G = np.array([ [np.dot(a1,a2) for a1 in lattice_vectors] for a2 in lattice_vectors ])
    #Metric tensor in reciprocal space
    if reciprocal_space: return np.linalg.inv(G)
    else: return G



from yambopy import *
from yambopy.plot  import *

# read the yambo_nl databases
NLDB=YamboNLDB()

#store the polarization in the pol array (it contains the three Cartesian directions)
pol =NLDB.Polarization[0]

#the time series of the simulation
time=NLDB.IO_TIME_points

#time when the external field was turned on
t_initial=NLDB.Efield[0]["initial_time"]

# create an empty array for the damped polarization
pol_damped=np.empty_like(pol)

# damp the polarization using a Lorentzian function for all the three directions
for i_d in range(3):
    pol_damped[i_d,:]=damp_it(pol[i_d,:],time,t_initial,damp_type='LORENTZIAN',damp_factor=0.1/ha2ev)

# Plot the polarization
Plot_Pol(time=time, pol=pol_damped, xlim=[0,55], save_file='polarization.pdf')

# Fourier transfort the polarization, divide by the external field to get \xhi(\omega) = P(\omega)/E(\omega)
# and then get \epsilon=1 + 4 \pi \xhi
Linear_Response(time=time,pol=pol_damped,efield=NLDB.Efield[0],plot=False,plot_file='eps.pdf')



# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
"""
This file contains classes and methods to generate input files

The FiniteDifferencesPhononFlow
"""

from qepy.pw import PwIn
from qepy.ph import PhIn
from qepy.pwxml import PwXML
from qepy.lattice import *
from qepy.matdyn import Matdyn
from qepy import qepyenv
from yambopy.io.inputfile import YamboIn
from yambopy.tools.duck import isiter
from yambopy.flow import PwTask, PhTask, P2yTask, YamboTask, DynmatTask, YambopyFlow, AbinitTask, E2yTask

__all__ = [
"FiniteDifferencesPhononFlow",
"KpointsConvergenceFlow",
"PwNscfYamboIPChiTasks",
"YamboQPBSETasks",
"PwNscfTasks",
"PwBandsTasks",
"PwRelaxTasks",
"PhPhononTasks"
]

class FiniteDifferencesPhononFlow():
    """
    This class takes as an input one structure and a phonon calculation.
    It produces a flow with the QE input files displaced along the phonon modes

    Author: Henrique Miranda
    """
    def __init__(self,structure,phonon_modes):
        self.structure = structure
        if not isinstance(phonon_modes,Matdyn):
            raise ValueError('phonon_modes must be an instance of Matdyn')
        self.phonon_modes = phonon_modes

    def get_tasks(self,path,kpoints,ecut,nscf_bands,nscf_kpoints=None,
                  imodes_list=None,displacements=[0.01,-0.01],iqpoint=0,**kwargs):
        """
        Create a flow with all the tasks to perform the calculation
        
        Arguments:
            generator: a builder function that takes structure, kpoints, ecut, 
                       nscf_bands, nscf_kpoints and kwargs as argument and returns the
                       tasks to be performed at each displacement
        """

        if imodes_list is None: imodes_list = list(range(self.phonon_modes.nmodes))
        if not isiter(displacements): displacements = [displacements]
        generator = kwargs.pop("generator",PwNscfYamboIPChiTasks)
             
        #create qe input from structure
        pwin = PwIn.from_structure_dict(self.structure,kpoints=kpoints,ecut=ecut)

        tasks = []
        #
        # Undisplaced structure
        #
        tasks.extend(generator(pwin.get_structure(),kpoints,ecut,nscf_bands,nscf_kpoints=nscf_kpoints,**kwargs))

        #
        # Displaced structures
        #
        reference = [dict(imode=None,displacement=0)]

        for imode in imodes_list:
            #get phonon mode
            cart_mode = self.phonon_modes.modes[iqpoint,imode]
            #iterate over displacements
            for displacement in displacements:
                #displace structure
                input_mock = pwin.get_displaced(cart_mode, displacement=displacement)
                displaced_structure = input_mock.get_structure()

                #generate tasks
                new_tasks = generator(displaced_structure,kpoints,ecut,
                                      nscf_bands,nscf_kpoints=nscf_kpoints,**kwargs)
                #save task
                tasks.extend(new_tasks)
                reference.append([dict(imode=None,displacement=displacement)])

        #store a reference for each of the tasks
        self.reference = reference
        return tasks

    
    def get_flow(self,path,kpoints,ecut,nscf_bands,nscf_kpoints=None,imodes_list=None,**kwargs):

        tasks = self.get_tasks(path=path,kpoints=kpoints,ecut=ecut,nscf_bands=nscf_bands,
                               nscf_kpoints=nscf_kpoints,imodes_list=imodes_list,**kwargs)
       
        #put all the tasks in a flow
        self.yambo_flow = YambopyFlow.from_tasks(path,tasks)
        return self.yambo_flow

    def get_dchi(self):
        """Collect all tasks producing eps files"""
        raise NotImplementedError('TODO')
        #find all the optical tasks
        #calculate finite differences
        pass

    def plot_ax(self,what):
        """ Collect and plot the results
        
        Arguments:
            what: what to plot can be eps for optical response
        """
        raise NotImplementedError('TODO')
        #plot finite differences
        pass        

    @property
    def path(self):
        return self.yambo_flow.path

class KpointsConvergenceFlow():
    """
    This class takes as an input one structure.
    It produces a flow with the QE input files with different number of kpoints
    and then runs a task staring from the QE input

    Author: Henrique Miranda
    """
    def __init__(self,structure):
        self.structure = structure
    
    def get_tasks(self,scf_kpoints,nscf_kpoints_list,ecut,nscf_bands,**kwargs):
        """
        Create a flow with all the tasks to perform the calculation
        
        Arguments:
            generator: a builder function that takes structure, kpoints, ecut, 
                       nscf_bands, nscf_kpoints and kwargs as argument and returns the
                       tasks to be performed at each displacement
        """
        generator = kwargs.pop("generator",PwNscfYamboIPChiTasks)

        #create a QE scf task and run
        qe_input = PwIn.from_structure_dict(self.structure,kpoints=scf_kpoints,ecut=ecut)
        qe_scf_task = PwTask.from_input(qe_input)
        tasks = [qe_scf_task]

        for nscf_kpoints in nscf_kpoints_list:

            #generate tasks
            new_tasks = generator(structure=self.structure,kpoints=scf_kpoints,
                                  nscf_kpoints=nscf_kpoints,
                                  ecut=ecut,nscf_bands=nscf_bands,**kwargs)
            tasks.extend(new_tasks)

        return tasks

    def get_flow(self,path,scf_kpoints,nscf_kpoints_list,ecut,nscf_bands,**kwargs):
        tasks = self.get_tasks(scf_kpoints=scf_kpoints,ecut=ecut,nscf_bands=nscf_bands,
                               nscf_kpoints_list=nscf_kpoints_list,**kwargs)
       
        #put all the tasks in a flow
        self.yambo_flow = YambopyFlow.from_tasks(path,tasks,**kwargs)
        return self.yambo_flow

class BandsConvergenceFlow():
    """
    This class takes as an input one structure.
    It produces a flow with the QE input files for nscf calculation with the max number of bands
    Then adds one yambo task with different number of bands

    Author: Henrique Miranda
    """
    def __init__(self,structure):
        self.structure = structure
    
    def get_tasks(self,scf_kpoints,ecut,nscf_kpoints,bands_list,**kwargs):
        """
        Create a flow with all the tasks to perform the calculation
        
        Arguments:
            generator: a builder function that takes structure, kpoints, ecut, 
                       nscf_bands, nscf_kpoints and kwargs as argument and returns the
                       tasks to be performed at each displacement
        """
        generator = kwargs.pop("generator",YamboIPChiTask)
        conv_thr = kwargs.pop("conv_thr",qepyenv.CONV_THR)

        #create a QE scf task and run
        nscf_bands = max(bands_list)
        tasks = PwNscfTasks(self.structure,kpoints=scf_kpoints,ecut=ecut,
                            nscf_bands=nscf_bands,nscf_kpoints=nscf_kpoints,conv_thr=conv_thr)
        qe_scf_task, qe_nscf_task, p2y_task = tasks
        tasks = list(tasks)

        link_task = p2y_task
        new_task  = p2y_task
        #we invert the list and compute from higher bands to lower
        #this avoids recomputing the dipoles
        for bands in sorted(bands_list,reverse=True):
            #generate tasks
            new_task = generator(link_task,bands=bands,dependencies=new_task,**kwargs)
            tasks.append(new_task)
            link_task = [p2y_task,new_task]

        return tasks

    def get_flow(self,path,scf_kpoints,ecut,nscf_kpoints,bands_list,**kwargs):
        tasks = self.get_tasks(scf_kpoints=scf_kpoints,ecut=ecut,nscf_kpoints=nscf_kpoints,
                               bands_list=bands_list,**kwargs)
       
        #put all the tasks in a flow
        self.yambo_flow = YambopyFlow.from_tasks(path,tasks,**kwargs)
        return self.yambo_flow


#inject some code in the initialization
def copy_pp(self):
    """
    This code is to be executed during the initialization of the SOC calculation
    It uses the dielectric function calculated without spin orbit coupling to
    calculate the QP corrections with spin-orbit coupling
    """
    import os
    import glob
    import shutil
    from netCDF4 import Dataset
    nospin_task = self.get_vars('nospin_task')
    spin_bands = self.get_vars('spin_bands')
    src = os.path.join(nospin_task.path,'run')
    dst = os.path.join(self.path,'run')
    if not os.path.isdir(dst): os.mkdir(dst)

    #link all the pp files
    ppfiles = glob.glob(os.path.join(src,'ndb.pp*'))
    for ppfile_path in ppfiles:
        ppfile = os.path.basename(ppfile_path)
        shutil.copy(ppfile_path,os.path.join(dst,ppfile))

    #change the spin value in the main file
    #also need to change the number of bands so yambo will read it
    main_pp = os.path.join(dst,'ndb.pp')
    with Dataset(main_pp,'r+') as dst_db:
        dst_db.variables['SPIN_VARS'][1]=2
        dst_db.variables['X_PARS_1'][2]=spin_bands

def update_cell_and_positions(self):
    """
    Code executed after the lattice relaxation to read the
    resulting cell parameters
    """
    from qepy import PwXML
    pwtask = self.get_vars("pwtask")
    prefix = pwtask.pwinput.prefix
    pwxml = PwXML(prefix,path=pwtask.path)
    self.pwinput.update_structure_from_xml(pwxml)

def get_scissor(self):
    """
    This code is to be executed during the intialization of a BSE calculation
    with scissor operator. It computes the scissor shift from a QP database
    and sets the BSE input accordingly
    """
    import os
    from yambopy.dbs.qpdb import YamboQPDB
    qp_task = self.get_vars('qp_task')
    valence = self.get_vars('valence')
    qp_path = os.path.join(qp_task.path,'run')
    #read qp database
    qpdb = YamboQPDB.from_db(folder=qp_path)
    scissor = qpdb.get_scissor(valence,verbose=0)[:3]
    #set scissor in the current input file
    self.yamboin_dict['KfnQP_E'] = list(scissor)

class SpinOrbitFlow():
    """
    Perform a yambo calculation including spin-orbit coupling.
    The dielectric function is calculated from the system without spin-orbit
    This has been tested for GW and BSE

    Author: Henrique Miranda
    """
    def __init__(self,structure,structure_spin=None):
        self.structure_nospin = structure
        if structure_spin is None: self.structure_spin = structure
  
    def get_tasks(self,scf_kpoints,ecut,nscf_kpoints,chi_bands,spin_bands,**kwargs):
        """
        Get a list of tasks executing this flow
        """
        tasks=[]
        conv_thr = kwargs.pop("conv_thr",qepyenv.CONV_THR)

        #create a yambo qp run
        yamboin_default_dict = dict(BndsRnXp=[1,chi_bands],
                                    NGsBlkXp=[1,'Ry'],
                                    EXXRLvcs=[10,'Ry'],
                                    QPkrange=[1,1,1,spin_bands],
                                    GbndRnge=[1,spin_bands])

        yamboin_dict = kwargs.pop("yamboin_dict",yamboin_default_dict)
        generator = kwargs.pop("generator",YamboQPTask)
        pp_runlevel = kwargs.pop("pp_runlevel",'-p p -V all')
        qp_runlevel = kwargs.pop("qp_runlevel",'-p p -g n -V all')
        spin_runlevel = kwargs.pop("spin_runlevel",qp_runlevel) 

        #without spin
        new_tasks = PwNscfTasks(self.structure_nospin,kpoints=scf_kpoints,ecut=ecut,
                            nscf_bands=chi_bands,nscf_kpoints=nscf_kpoints,conv_thr=conv_thr)
        qe_scf_task, qe_nscf_task, p2y_task = new_tasks

        nospin_task = generator(p2y_task,runlevel=pp_runlevel,yamboin_dict=yamboin_dict,dependencies=p2y_task,**kwargs)

        tasks.extend( [qe_scf_task, qe_nscf_task, p2y_task, nospin_task] )

        #with spin
        new_tasks = PwNscfTasks(self.structure_spin,kpoints=scf_kpoints,ecut=ecut,
                            nscf_bands=spin_bands,nscf_kpoints=nscf_kpoints,conv_thr=conv_thr)
        qe_scf_task, qe_nscf_task, p2y_task = new_tasks

        #patch the input files to include spin-orbit coupling
        qe_scf_task.get_instances_from_inputs(PwIn)[0].set_spinorbit()
        qe_nscf_task.get_instances_from_inputs(PwIn)[0].set_spinorbit()

        spin_task = generator(p2y_task,runlevel=spin_runlevel,yamboin_dict=yamboin_dict,dependencies=p2y_task,**kwargs)

        tasks.extend( [qe_scf_task, qe_nscf_task, p2y_task, spin_task] )

        spin_task.set_vars('spin_bands',spin_bands)
        spin_task.set_vars('nospin_task',nospin_task)
        spin_task.set_code('initialize',copy_pp)

        return tasks

    def get_flow(self,path,scf_kpoints,ecut,nscf_kpoints,chi_bands,spin_bands,**kwargs):
        tasks = self.get_tasks(scf_kpoints=scf_kpoints,ecut=ecut,nscf_kpoints=nscf_kpoints,
                               chi_bands=chi_bands,spin_bands=spin_bands,**kwargs)
       
        #put all the tasks in a flow
        self.yambo_flow = YambopyFlow.from_tasks(path,tasks,**kwargs)
        return self.yambo_flow

def PwNscfYamboIPChiTasks(structure,kpoints,ecut,nscf_bands,
                          yambo_runlevel='-o c -V all',nscf_kpoints=None,**kwargs):
    """
    Return the PwNscfTasks and a YamboTask for and IP calculation
    """ 
    conv_thr = kwargs.pop("conv_thr",qepyenv.CONV_THR)

    #create scf, nscf and p2y task
    tmp_tasks = PwNscfTasks(structure,kpoints,ecut,nscf_bands,nscf_kpoints,conv_thr=conv_thr)
    qe_scf_task,qe_nscf_task,p2y_task = tmp_tasks

    yambo_task = YamboIPChiTask(p2y_task,yambo_runlevel=yambo_runlevel,**kwargs)

    return qe_scf_task,qe_nscf_task,p2y_task,yambo_task

def YamboIPChiTask(p2y_task,**kwargs):
    """
    Return a yambo IP task
    """
    yambo_ip_default_dict = dict(QpntsRXd=[1,1],
                                 ETStpsXd=1000)
    yambo_ip_dict = kwargs.pop('yambo_ip_dict',yambo_ip_default_dict)
    yambo_runlevel = kwargs.pop('yambo_runlevel','-o c -V all')
    bands = kwargs.pop('bands',None)
    dependencies = kwargs.pop('dependencies',p2y_task)
    if bands: yambo_ip_dict['BndsRnXd'] = [1,bands]
    
    #add yambo_task
    yambo_task = YamboTask.from_runlevel(p2y_task,yambo_runlevel,yambo_ip_dict,
                                         dependencies=dependencies)
    return yambo_task

def YamboQPBSETasks(p2y_task,qp_dict,bse_dict,qp_runlevel='-p p -g n -V all',
                    bse_runlevel='-p p -k sex -y d -V all',valence=False,**kwargs):
    """
    Return a QP and BSE calculation
    """
    import copy
    dependencies = kwargs.pop('dependencies',p2y_task)
    inputs = [p2y_task]
    additional_inputs = kwargs.pop('additional_inputs',None)
    if additional_inputs: 
        if not isiter(additional_inputs): additional_inputs = [additional_inputs]
        inputs.extend(additional_inputs)

    #create a yambo qp run
    qp_task = YamboTask.from_runlevel(inputs,qp_runlevel,qp_dict,dependencies=dependencies)

    #create a yambo qp+bse run
    bse_dict['KfnQPdb']="E < run/ndb.QP"
    qp_bse_task = YamboTask.from_runlevel([p2y_task,qp_task],bse_runlevel,bse_dict,
                                          dependencies=qp_task)
    
    #create a yambo scissor+bse run
    if valence:
        bse_dict = copy.deepcopy(bse_dict)
        bse_dict.pop('KfnQPdb')
        scissor_bse_task = YamboTask.from_runlevel([p2y_task,qp_task],bse_runlevel,bse_dict,
                                                   dependencies=qp_task)
        scissor_bse_task.set_vars('qp_task',qp_task)
        scissor_bse_task.set_vars('valence',valence)
        scissor_bse_task.set_code('initialize',get_scissor)
        return qp_task, qp_bse_task, scissor_bse_task
    
    return qp_task, qp_bse_task


def YamboQPTask(p2y_task,yamboin_dict,runlevel='-p p -g n -V all',**kwargs):
    """
    Return a QP calculation
    """
    #create a yambo qp run
    dependencies = kwargs.pop('dependencies',p2y_task)
    qp_task = YamboTask.from_runlevel(p2y_task,runlevel,yamboin_dict,dependencies=p2y_task)

    return qp_task

def PhPhononTasks(structure,kpoints,ecut,qpoints=None):
    """
    Return a ScfTask, a PhTask and Matdyn task
    """

    #create a QE scf task and run
    qe_input = PwIn.from_structure_dict(structure,kpoints=kpoints,ecut=ecut)
    qe_scf_task = PwTask.from_input(qe_input)

    #create phonon tasks
    if qpoints is None: qpoints = qe_input.kpoints
    ph_input = PhIn.from_qpoints(qpoints)
    ph_task = PhTask.from_scf_task([ph_input,qe_scf_task],dependencies=qe_scf_task)

    #create matdyn task
    matdyn_task = DynmatTask.from_phonon_task(ph_task,dependencies=ph_task)

    return qe_scf_task, ph_task, matdyn_task

def PwNscfTasks(structure,kpoints,ecut,nscf_bands,nscf_kpoints=None,**kwargs):
    """
    Return a ScfTask, NscfTask and P2yTask preparing for a Yambo calculation
    """
    scf_conv_thr = kwargs.pop("conv_thr",qepyenv.CONV_THR)
    scf_conv_thr = kwargs.pop("scf_conv_thr",scf_conv_thr)
    nscf_conv_thr = kwargs.pop("nscf_conv_thr",scf_conv_thr*10)
    nscf_paralelization = kwargs.pop("nscf_paralelization","")

    #create a QE scf task and run
    qe_input = PwIn.from_structure_dict(structure,kpoints=kpoints,ecut=ecut,conv_thr=scf_conv_thr)
    qe_scf_task = PwTask.from_input(qe_input)

    # pseudo_dir
    pseudo_dir = kwargs.pop("pseudo_dir", None)
    if pseudo_dir: qe_input.control['pseudo_dir'] = "'%s'" % pseudo_dir

    #Spin
    spin = kwargs.pop("spin", None)
    if spin=="spinor": qe_input.set_spinorbit()
    if spin=="polarized": qe_input.set_spinpolarized() 

    #Magnetization
    starting_magnetization = kwargs.pop("starting_magnetization", None)
    qe_input.set_magnetization(starting_magnetization)

    #create a QE nscf task and run
    if nscf_kpoints is None: nscf_kpoints = kpoints
    qe_input = qe_input.copy().set_nscf(nscf_bands,nscf_kpoints,conv_thr=nscf_conv_thr)
    qe_nscf_task = PwTask.from_input([qe_input,qe_scf_task],dependencies=qe_scf_task,paralelization=nscf_paralelization)

    #create a p2y nscf task and run
    p2y_task = P2yTask.from_nscf_task(qe_nscf_task)

    return qe_scf_task, qe_nscf_task, p2y_task

def PwBandsTasks(structure,kpoints,ecut,nscf_bands,path_kpoints,**kwargs):
    """
    Return a ScfTask and BandsTask (Author: AMS, HM)

    Args:
        kpoints: the k-mesh for the scf calculation
        ecut: the planewave basis cutoff
        nscf_bands: number of bands in the nscf_calculation
        path_kpoints: the k-points along the path
        spin: can be 'polarized' for calculation with spin or 'spinor' for calculation with spin-orbit
        starting_magnetization: a list with the starting magnetizations for each atomic type
    """
    scf_conv_thr = kwargs.pop("conv_thr",qepyenv.CONV_THR)
    scf_conv_thr = kwargs.pop("scf_conv_thr",scf_conv_thr)
    bands_conv_thr = kwargs.pop("nscf_conv_thr",scf_conv_thr*10)

    #create a QE scf task
    qe_input = PwIn.from_structure_dict(structure,kpoints=kpoints,ecut=ecut,conv_thr=scf_conv_thr)

    # pseudo_dir
    pseudo_dir = kwargs.pop("pseudo_dir", None)
    if pseudo_dir: qe_input.control['pseudo_dir'] = "'%s'" % pseudo_dir

    #Spin
    spin = kwargs.pop("spin", None)
    if spin=="spinor": qe_input.set_spinorbit()
    if spin=="polarized": qe_input.set_spinpolarized() 

    #Magnetization
    starting_magnetization = kwargs.pop("starting_magnetization", None)
    qe_input.set_magnetization(starting_magnetization)

    pseudo_dir = kwargs.pop("pseudo_dir", None)
    if pseudo_dir: qe_input.control['pseudo_dir'] = "'%s'" % pseudo_dir

    scf_paralelization = kwargs.pop("scf_paralelization","")
    qe_scf_task = PwTask.from_input(qe_input,paralelization=scf_paralelization)

    #create a QE bands task
    qe_input_bands = qe_input.copy().set_bands(nscf_bands,path_kpoints=path_kpoints,conv_thr=bands_conv_thr,**kwargs)
    nscf_paralelization = kwargs.pop("nscf_paralelization","")
    qe_bands_task = PwTask.from_input([qe_input_bands,qe_scf_task],dependencies=qe_scf_task,paralelization=nscf_paralelization,**kwargs)

    return qe_scf_task, qe_bands_task

def PwRelaxTasks(structure,kpoints,ecut,cell_dofree='all',**kwargs):
    """
    Return a RelaxTask and ScfTask (Author: AMS, HM)

    Args:
        kpoints: the k-mesh for the scf calculation
        ecut: the planewave basis cutoff
        cell_free
        spin: can be 'polarized' for calculation with spin or 'spinor' for calculation with spin-orbit
        starting_magnetization: a list with the starting magnetizations for each atomic type
    """
    #raise NotImplementedError('This function is under development!')

    scf_conv_thr = kwargs.pop("conv_thr",qepyenv.CONV_THR)
    scf_conv_thr = kwargs.pop("scf_conv_thr",scf_conv_thr)
    paralelization = kwargs.pop("paralelization","")

    #create a QE input scf
    qe_input_scf = PwIn.from_structure_dict(structure,kpoints=kpoints,ecut=ecut,conv_thr=scf_conv_thr)

    #Spin
    spin = kwargs.pop("spin", None)
    if spin=="spinor": qe_input_scf.set_spinorbit()
    if spin=="polarized": qe_input_scf.set_spinpolarized() 

    #Magnetization
    starting_magnetization = kwargs.pop("starting_magnetization", None)
    qe_input_scf.set_magnetization(starting_magnetization)

    pseudo_dir = kwargs.pop("pseudo_dir", None)
    if pseudo_dir: qe_input_scf.control['pseudo_dir'] = "'%s'" % pseudo_dir

    #create a QE relax-atom task
    qe_input_relax_atoms = qe_input_scf.copy().set_relax(cell_dofree=None)
    qe_relax_atoms_task = PwTask.from_input(qe_input_relax_atoms,paralelization=paralelization)

    #create a QE relax-cell task
    qe_input_relax_cell = qe_input_scf.copy().set_relax(cell_dofree=cell_dofree)
    qe_relax_cell_task = PwTask.from_input([qe_input_relax_cell,qe_input_relax_atoms],dependencies=qe_relax_atoms_task,paralelization=paralelization)
    qe_relax_cell_task.set_vars("pwtask",qe_relax_atoms_task)
    qe_relax_cell_task.set_code("initialize",update_cell_and_positions)

    #create a QE scf task
    qe_scf_task = PwTask.from_input([qe_input_scf,qe_input_relax_cell],dependencies=qe_relax_cell_task,paralelization=paralelization)
    qe_scf_task.set_vars("pwtask",qe_relax_cell_task)
    qe_scf_task.set_code("initialize",update_cell_and_positions)

    return qe_relax_atoms_task, qe_relax_cell_task, qe_scf_task

def AbinitNscfTasks(structure,kpoints,ecut,nscf_bands,nscf_kpoints=None,**kwargs):
    from abipy.core.structure import Structure
    from abipy.abio.factories import scf_for_phonons
    from pymatgen.core.units import bohr_to_ang

    #extract pseudos
    pseudo_list = []
    for atype,(mass,pseudo) in structure['atypes'].items():
        pseudo_list.append(pseudo)
    pseudo_table = kwargs.pop("pseudo_table",pseudo_list)
         
    #create a PwInput file just to read the ibrav from structure
    qe_input = PwIn.from_structure_dict(structure)
    lattice,coords,species = qe_input.get_cell()
    lattice = [[col*bohr_to_ang for col in row] for row in lattice]
    structure = Structure(lattice,species,coords)

    #create an AbinitInput file from structure
    spin_mode = kwargs.pop('spin_mode','unpolarized')
    smearing = kwargs.pop('smearing','nosmearing')
    inp = scf_for_phonons(structure, pseudo_table, spin_mode=spin_mode, 
                          smearing=smearing, ecut=ecut/2)

    return AbinitNscfTasksFromAbinitInput(inp,kpoints,ecut,nscf_bands,nscf_kpoints=nscf_kpoints,**kwargs)
 
def AbinitNscfTasksFromAbinitInput(inp,kpoints,ecut,nscf_bands,nscf_kpoints=None,**kwargs):
    tolwfr = kwargs.pop("tolwfr",1e-22)
    nbdbuf = kwargs.pop("nbdbuf",int(nscf_bands*0.15))   
   
    #scf
    inp = inp.deepcopy()
    inp.set_kmesh(ngkpt=kpoints,
                  kptopt=1,
                  shiftk=(0,0,0))
    abinit_scf_task = AbinitTask.from_input(inp)

    #nscf
    inp = inp.deepcopy()
    if nscf_kpoints is None: nscf_kpoints = kpoints 
    inp.set_kmesh(ngkpt=nscf_kpoints,
                  kptopt=1,
                  shiftk=(0,0,0))
    inp.set_vars(iscf=-2,
                 getden=1,
                 nband=nscf_bands+nbdbuf,
                 nbdbuf=nbdbuf)
    abinit_nscf_task = AbinitTask.from_input([inp,abinit_scf_task],dependencies=abinit_scf_task)

    #e2y task
    e2y_task = E2yTask.from_nscf_task(abinit_nscf_task)

    return abinit_scf_task, abinit_nscf_task, e2y_task



# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#

"""
This file contains examples of structures
"""

__all__ = [
    "BN",
    "MoS2",
    "Si"
]

#BN
a = 4.7
c = 12
lattice = dict(ibrav=4,celldm1=4.7,celldm3=c/a)
atypes = dict(B=[10.811, "B.pbe-mt_fhi.UPF"],
              N=[14.0067,"N.pbe-mt_fhi.UPF"])

atoms = [['N',[ 0.0, 0.0,0.5]],
         ['B',[1./3,2./3,0.5]]]
BN = dict(lattice=lattice,atypes=atypes,atoms=atoms)

#MoS2
a = 5.838
c = 18
lattice = dict(ibrav=4,celldm1=5.838,celldm3=c/a) 
atypes = dict(Mo=[95.94, "Mo.pz-mt_fhi.UPF"],
              S =[32.065, "S.pz-mt_fhi.UPF"]) 
atoms = [['Mo',[2./3,1./3,          0.0]],
         [ 'S',[1./3,2./3, 2.92781466/c]],
         [ 'S',[1./3,2./3,-2.92781466/c]]]
MoS2 = dict(lattice=lattice,atypes=atypes,atoms=atoms)

#Si
lattice = dict(ibrav=2,celldm1=10.3)
atypes = dict(Si=[28.086,"Si.pbe-mt_fhi.UPF"])
atoms = [['Si',[0.125,0.125,0.125]],
         ['Si',[-.125,-.125,-.125]]]
Si = dict(lattice=lattice,atypes=atypes,atoms=atoms)




import sys
import argparse
import numpy as np
import math
from scipy import special
from yambopy.units import ha2ev,fs2aut,AU2KWCMm2

"""
In this example we show how to generate and external field
in a format that readable from yambo_rt and yambo_nl.
We give two example SIN and GAUSS functions

An external field should provide f(t), f'(t), f''(t) 
I put all paramters of the field inside their repecitve functions

*** IMPORTANT ***
The field is save without field intensity(amplitude) this is read
from the yambo_nl/yambo_rt input
"""

def sin_field(t_steps):
    # In this example we generate a field with frequency 2.0 eV
    freq=2.0/ha2ev 
    f_t   = -1.0/freq*(np.cos(freq*t_steps)-1.0)   # A(t) 
    fp_t  = np.sin(freq*t_steps)                  # A'(t)
    fpp_t = freq*np.cos(freq*t_steps)             # A''(t)
    return  f_t,fp_t,fpp_t 


def gauss_field(t_steps):
    # In this example we generate a guassian external field
    sigma=5.0*fs2aut # Field width
    T_0  =3.0*sigma  
    Expf =np.exp(-(t_steps-T_0)**2/(2.0*sigma**2) )
    f_t  =sigma*np.sqrt(math.pi/2.0)*(special.erf((t_steps-T_0)/(sigma*np.sqrt(2.0)))+1.0)
    fp_t =Expf
    fpp_t=-Expf*(t_steps-T_0)/sigma**2
    return  f_t,fp_t,fpp_t 

    

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='External field generation',formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-d', '--tstep',  type=float, default=0.01,      help='Time step (fs)')
    parser.add_argument('-t', '--trange', type=float, nargs=2, default=[0.0,80.0],help='Time range (fs)')
    parser.add_argument('-s', '--tstart', type=float, default=0.01,      help='Initial time (fs)')
    parser.add_argument('-v', '--versor', type=float, nargs=3, default=[1.0,0.0,0.0],help='Field versor')
    parser.add_argument('-f', '--fname',  type=str, help='Field name (SIN | GAUSS) ')
    parser.add_argument('-o', '--fout',  type=str, default="EXTFIELD1_P1.time",help='External field file name')
    args = parser.parse_args()

    if len(sys.argv)==1:
        parser.print_help()
        sys.exit(1)

    if args.fname == None:
        print("You have to specify the field name ")
        sys.exit(1)

    t_start=args.tstart*fs2aut
    t_step =args.tstep*fs2aut
    t_range=np.arange(args.trange[0]*fs2aut,args.trange[1]*fs2aut+t_step/2.0,t_step)
    
    print("\n\n * * * Generate and external field for yambo_rt/yambo_nl * * * \n\n")
    print("Field name : ",args.fname)
    print("Time range : ",args.trange,"[fs]")
    print("Time step  : ",args.tstep, "[fs]")
    print("Start time : ",args.tstart,"[fs]")

    if args.fname == "SIN":
        a_pot=sin_field(t_range-t_start)
    elif args.fname == "GAUSS":
        a_pot=gauss_field(t_range-t_start)
    else:
        print("Unknown external field ")
        sys.exit(1)

    data = np.column_stack((t_range/fs2aut,a_pot[0],a_pot[1],a_pot[2]))
    with open(args.fout, "w") as f:
        f.write(str(t_range.size)+"     "+str(t_step)+" \n")
        np.savetxt(f, data, fmt="%4.8e", delimiter="\t")



class yambopyenv():
    YAMBO = "yambo"
    P2Y = "p2y"
    E2Y = "e2y"
    YPP = "ypp"
    SCHEDULER = "bash"
    YAMBO_RT = "yambo_rt"
    YPP_RT = "ypp_rt"
    YAMBO_NL = "yambo_nl"
    YPP_NL = "ypp_nl"


from __future__ import print_function
##############################################################################
#
# Author: Alejandro Molina-Sanchez
# Run real-time simulations with Yambo
#
# Warning: Real-time simulations requires several data folders for running
# properly. Before using these scripts compulsively, it is recommended
# to understand the different run levels.
#
# Instructions:
# The dictionary 'job' stores the variable of the calculation
# calculation : 'collisions', 'pump', 'dissipation'
# folder-col  : collision data
# folder-run  : results (only work if collisions have been previously calculated)
# DG          : True or False if we use the double-grid
# nodes       : cpu-dependent variable
# cores       :          "
#
##############################################################################
#from __future__ import print_function
from builtins import str
from yambopy     import *
from schedulerpy import *
import argparse

# Select the run-level : 'collision', 'pump', 'dissipation'
parser = argparse.ArgumentParser(description='Real-time simulation')
parser.add_argument('-c' ,'--collisions' ,action="store_true")
parser.add_argument('-p' ,'--pump'       ,action="store_true")
parser.add_argument('-d' ,'--dissipation',action="store_true")
args = parser.parse_args()

yambo_rt = 'yambo_rt'
folder   = 'rt-6x6'

# Generation of the input file

#os.system('cd %s; mkdir -p inputs' % folder)

if args.collisions:
  print('Collisions')
  run = YamboIn('%s -r -e -v hsex'%yambo_rt,folder=folder)
elif args.pump:
  print('Time-dependent with electric field')
  run = YamboIn('%s -q p'%yambo_rt,folder=folder)
elif args.dissipation:
  print('Time-dependent with electric field and electron-phonon scattering')
  run = YamboIn('%s -s p -q p'%yambo_rt,folder=folder)
else:
  print('Invalid calculation type')
  exit()

# Proportionality constant of the DOS(T)
def dos_temperature(temperature):
  ac, bc, cc = 0.506760, 0.000069, 0.000002
  av, bv, cv = 0.437618, 0.000070, 0.000002
  dos_c = ac + bc*temperature + cc*temperature**2
  dos_v = av + bv*temperature + cv*temperature**2
  return [dos_c,dos_v]

job = dict()
job['radiative']    = False
job['folder-col']   = 'col-hxc' # name of collisions folder (path folder/folder-col)
# the input file and run-folder names are defined at the end
job['folder-gkkp']  = 'gkkp'
job['DG']           = (False,'dg-60x60') # Double-grid (True|False) and DG folder
job['nodes']        = 1
job['cores']        = 12
job['ppn']          = job['cores']
job['threads']      = job['cores']*job['nodes']
job['name']         = 'MoS2-dynamics'
job['yambo_module'] = 'yambo/master-intel'
job['temperature']  = 0

# System Common variables
run['DBsIOoff']      = "GF P J"
#run['DBsIOoff']      = "GF CARRIERs"
run['RT_CPU']        = "%s.1.1.1"  % (job['nodes']*job['cores']) # [PARALLEL] CPUs for each role
run['RT_ROLEs']      = "k.b.q.qp"    # [PARALLEL] CPUs roles (k,b,q,qp)
run['HXC_Potential'] = "HARTREE+SEX" # [SC] SC HXC Potential
#run['RT_Threads']    = 1

# Collision variables
if args.collisions:
  run['FFTGvecs']      = [20  ,'Ha']
  run['EXXRLvcs']      = [1000,'mHa']  # Equivalent to BSENGBlk in the BSE run-level
  run['HARRLvcs']      = [20  ,'Ha']   # Equivalent to BSENGexx in the BSE run-level
  run['CORRLvcs']      = [1000,'mHa']  #
  run['COLLBands']     = [25,28]
  run['NGsBlkXs']      = [1000,'mHa']
  run['BndsRnXs']      = [1, 60]
  run['RandQpts']      = 1000000         # Only in layers
  run['RandGvec']      = [ 1 , 'RL' ]
  run['CUTGeo']        = 'box z'         # Only in layers
  run['CUTBox']        = [0,0,38]
  run['X_all_q_CPU']   = "1.%d.1.1" % (job['nodes']*job['cores'])     # [PARALLEL] CPUs for each role
  run['X_all_q_ROLEs'] = "q.k.c.v"       # [PARALLEL] CPUs roles (q,k,c,v)
  #run.arguments.append('ALLGHAR')
  run.write('%s/collisions.in'%folder)

# Common time-dependent variable
if args.pump or args.dissipation:
  run['RTBands']    = [25,28]
  #run['GfnQP_Wv']   = [0.0,dos_temperature(job['temperature'])[1],0.00]  # Only for dissipation
  #run['GfnQP_Wc']   = [0.0,dos_temperature(job['temperature'])[0],0.00]
  run['GfnQP_Wv']   = [0.025,0.0,0.0]
  run['GfnQP_Wc']   = [0.025,0.0,0.0]
  run['GfnQP_E']    = [1.04588, 1.00, 1.00]            # [EXTQP BSK BSS] E parameters  (c/v) eV|adim|adim
  # Time-propagation
  run['RTstep']     = [5.0,'as'] # Real-time step
  run['NETime']     = [600,'fs'] # Simulation duration
  run['Integrator'] = "RK2 RWA"
  run['IOtime']     = [ [ 50.0, 00.10, 1.00], 'fs' ]   # [RT] Time between to consecutive I/O (J,P,CARRIERs - GF - OUTPUT)
  #run['IOtime']     = [ [ 0.10, 0.10, 1.00], 'fs' ]   # [RT] Time between to consecutive I/O (J,P,CARRIERs - GF - OUTPUT)
  # Pump Pulse
  run['Field1_Int']      = [ 1E3 , 'kWLm2']    # Intensity pulse
  run['Field1_Dir']      = [0.0,1.0,0.0]       # Polarization pulse
  run['Field1_Dir_circ'] = [1.0,0.0,0.0]       # Polarization pulse (second axis for circular)
  run['Field1_pol']      = "linear"            # Polarization type (linear or circular)
  run['Field1_kind'] = "QSSIN"                 # [RT Field1] Kind(SIN|RES|ANTIRES|GAUSS|DELTA|QSSIN)
  run['Field1_Damp'] = [ 70.0,'fs']            # Damping (width of pulse)
  run['Field1_Freq'] = [[2.0,2.0],'eV']
  if job['radiative']:
    run.arguments.append('el_photon_scatt')

# Pumping with finite pulse and electron-phonon dissipation
if args.dissipation:
# Interpolation
  run['BoseTemp']        = [ job['temperature'], 'K']
  run['LifeExtrapSteps'] = [ [0.05,0.05], 'fs' ] # Extrapolation time for phonon-assisted dissipation
  run['ElPhModes']       = [ 1, 9]
  run.arguments.append('LifeExtrapolation')   # Commented:   Lifetimes are constant

# Run -- Creating

print('Collisions        ',job['folder-col'])
print('Number of nodes   ',job['nodes'])
print('Number of cores   ',job['cores'])

oarsub = oarsub(nodes=job['nodes'],core=job['cores'],dependent=0,name=job['name'],walltime="24:00:00")
oarsub.add_command('module load %s'%job['yambo_module'])
oarsub.add_command('export OMP_NUM_THREADS=1')

# Collisions
if args.collisions:
  oarsub.add_command('cd %s; mpirun -hostfile \$OAR_NODEFILE %s -F collisions.in -J %s -C %s'%(folder,yambo_rt,job['folder-col'],job['folder-col']))
  oarsub.write('%s/collisions.ll' % folder )
  oarsub.run()
  print('running yambo-collision')

# Time-dependent without dissipation
if args.pump:

  # name of run - change here if you want different variables
  if run['Field1_kind'] == 'DELTA':
    job['folder-run'] = 'DELTA-%.0e' % ( run['Field1_Int'][0])
  else:
    job['folder-run'] = '%s' %(run['Field1_kind'])
    job['folder-run'] += '-%.0e-%sfs-%seV-%sK' % ( run['Field1_Int'][0], run['Field1_Damp'][0], run['Field1_Freq'][0][0],job['temperature'])

  # writing input file
  os.system('mkdir -p %s/%s'%(folder,job['folder-run']))
  run.write('%s/%s/pulse.in'% (folder,job['folder-run']))

  # submission script
  oarsub.add_command('cd %s/%s; mpirun -hostfile \$OAR_NODEFILE %s -F pulse.in -J \'pulse,..//%s\' -C pulse -I \'../\''%(folder,job['folder-run'],yambo_rt,job['folder-col']) )
  oarsub.write('%s/%s/pulse.ll'%(folder,job['folder-run']))
  oarsub.run()
  print('running Dynamics without dissipation in folder: ' + str(job['folder-run']))

# Time-dependent with dissipation
if args.dissipation:

  # name of run - change here if you want different variables
  if run['Field1_kind'] == 'DELTA':
    job['folder-run'] = 'DELTA-D-%.0e' % ( run['Field1_Int'][0] )
  else:
    job['folder-run'] = '%s' %(run['Field1_kind'])
    job['folder-run'] += '-D-%seV-%sK-%sfs' % ( run['Field1_Freq'][0][0],job['temperature'],run['LifeExtrapSteps'][0][0])
    #job['folder-run'] += '-D-%.0e-%sfs-%seV-%sK' % ( run['Field1_Int'][0], run['Field1_Damp'][0], run['Field1_Freq'][0][0],job['temperature'])
  if job['DG'][0]:
    job['folder-run'] += '-DG'

  # writing input file
  os.system('mkdir -p %s/%s'%(folder,job['folder-run']))
  run.write('%s/%s/pulse.in'% (folder,job['folder-run']))

  # submission script
  if job['DG'][0]:
    oarsub.add_command('cd %s/%s; mpirun -hostfile \$OAR_NODEFILE %s -F pulse.in -J \'pulse,..//%s,..//%s,..//%s\' -C pulse -I \'../\''%(folder,job['folder-run'],yambo_rt,job['folder-col'],job['folder-gkkp'],job['DG'][1]) )
    print('Double Grid enabled')
  else:
    oarsub.add_command('cd %s/%s; mpirun -hostfile \$OAR_NODEFILE %s -F pulse.in -J \'pulse,..//%s,..//%s\' -C pulse -I \'../\''%(folder,job['folder-run'],yambo_rt,job['folder-col'],job['folder-gkkp']) )
  oarsub.write('%s/%s/pulse.ll'%(folder,job['folder-run']))
  oarsub.run()
  print('running Dynamics with dissipation in folder: ' + str(job['folder-run']))



from yambopy.tools.duck import isstring
import yamboparser

class YamboFile(yamboparser.YamboFile):
    """
    Wrapper around the Yambofile class of yamboparser
    """
    def from_dict(filename):
        """
        intialize from a dicitonary
        """
        pass
                
    def from_file(filename):
        """
        Read a file and find what type it is
        """
        #detect the type
        #store the data
        pass

    def write_json(self,filename):
        """
        Write a json file with the data for this file
        """
        pass

    @staticmethod
    def has_tag(filename,tags):
        """check if the filename has a tag in its name"""
        if isstring(tags):
            tags = (tags,)
        return any([tag in filename for tag in tags])

    @staticmethod
    def is_output(filename):
        """check if the file is output file"""
        return filename.startswith('o.') or filename.startswith('o-')

    @staticmethod
    def is_log(filename):
        """check if the file is log file"""
        return filename.startswith('l.') or filename.startswith('l-')

    @staticmethod
    def is_report(filename):
        """check if the file is report file"""
        return filename.startswith('r-')

class YamboGW(YamboFile):
    """
    Provide functions specific of GW calculations
    """
    def plot_gw(self):
        pass

class YamboEPS(YamboFile):
    """
    Provide functions specific of BSE calculations
    """
    def plot_eps(self):
        pass




"""Duck-typing tests (taken from abipy)"""

def isstring(s):
    """True if s behaves like a string (duck typing test)."""
    try:
        s + " "
        return True
    except TypeError:
        return False

def isiter(s):
    try:
        iterator = iter(s)
        return True
    except TypeError:
        return False



#
# Author: Henrique Pereira Coutada Miranda
# Run a BSE calculation using yambo
#
from __future__ import print_function
from builtins import range
import sys
from yambopy import *
from yambocommandline import*
from qepy import *
from schedulerpy import *
import argparse
import shutil
import matplotlib.pyplot as plt

yambo  = 'yambo'
p2y    = 'p2y'
prefix = 'bn'
folder = 'bse_conv'

scheduler = Scheduler.factory

def create_save():
    #check if the nscf cycle is present
    if os.path.isdir('nscf/%s.save'%prefix):
        print('nscf calculation found!')
    else:
        print('nscf calculation not found!')
        exit()

    #check if the SAVE folder is present
    if not os.path.isdir('database'):
        print('preparing yambo database')
        shell = scheduler()
        shell.add_command('cd nscf/%s.save; %s; %s'%(prefix,p2y,yambo))
        shell.add_command('cd ../../')
        shell.add_command('mkdir -p database')
        shell.add_command('mv nscf/%s.save/SAVE database'%prefix)
        shell.run()

    #create the folder to run the calculation
    if not os.path.isdir(folder):
        shell = scheduler()
        shell.add_command('mkdir -p %s'%folder)
        shell.add_command('cp -r database/SAVE %s/'%folder)
        shell.run()

def bse_convergence(what='screening',threads=1,nohup=False):
    if nohup: nohup = 'nohup'
    else:     nohup = ''

    #create the yambo input file
    y = YamboIn.from_runlevel('-X s -o b -k sex -y d -V all',executable=yambo,folder=folder)

    #default variables
    y['BSEBands'] = [4,5]
    y['BEnSteps'] = 500
    y['BSEEhEny'] = [[1.0,10.0],'eV']
    y['BEnRange'] = [[2.0,12.0],'eV']
    y['BSENGexx'] = [10,'Ry']
    y['KfnQP_E']  = [2.91355133,1.0,1.0] #some scissor shift
    y['DBsIOoff'] = 'BS' #turn off writting BSE hamiltonian DB (better performance)

    print(what)

    if what == 'screening':
        #list of variables to optimize the screening screening
        conv = { 'FFTGvecs': [[10,10,15,20,30],'Ry'],
                 'NGsBlkXs': [[1,1,2,3,5,6], 'Ry'],
                 'BndsRnXs': [[1,10],[1,10],[1,20],[1,30],[1,40]] }
    else:
        # converged parameters for epsilon
        y['FFTGvecs'] = [30,'Ry']
        y['NGsBlkXs'] = [2,'Ry']
        y['BndsRnXs'] = [[1,40],'Ry']

        # choose a large number of Bands in BSE
        # BSEEhEny will choose which transitions to include
        y['BSEBands'] = [1,10]

        #list of variables to optimize the BSE
        conv = { 'BSEEhEny': [[[1,10],[1,10],[1,12],[1,14]],'eV'],
                 'BSENGBlk': [[0,0,1,2], 'Ry'],
                 'BSENGexx': [[10,10,15,20],'Ry']}

    def run(filename):
        """
        Function to be called by the optimize function
        """
        path = filename.split('.')[0]
        print(filename, path)
        shell = scheduler()
        shell.add_command('cd %s'%folder)
        shell.add_command('%s mpirun -np %d %s -F %s -J %s -C %s 2> %s.log'%(nohup,threads,yambo,filename,path,path,path))
        shell.add_command('touch %s/done'%path)
        if not os.path.isfile("%s/%s/done"%(folder,path)):
            shell.run()

    y.optimize(conv,folder='bse_conv',run=run,ref_run=False)

def analyse():
    #pack the files in .json files
    pack_files_in_folder(folder)

    paths = []
    #get folder names
    for dirpath,dirnames,filenames in os.walk(folder):
        #ignore the root folder
        if dirpath == folder:
            continue

        #check if there are some output files in the folder
        if ([ f for f in filenames if 'o-' in f ]):
            paths.append( dirpath.split('/')[-1] )

    for path in paths:
        print( path )
        #get the absorption spectra
        a = YamboBSEAbsorptionSpectra(path)

    print( "To plot the data run:" )
    print( "python bse_conv_bn.py -p -e" )
    print( "python bse_conv_bn.py -p -b" )

def plot(what):
    #plot the results using yambo analyser
    y = YamboAnalyser(folder)
    print(y)

    fig = plt.figure(figsize=(10,8))
    if what == "screening":
        ax = plt.subplot(3,1,1)
        y.plot_bse(['eps','FFTGvecs'],ax=ax)
        ax = plt.subplot(3,1,2)
        y.plot_bse(['eps','NGsBlkXs'],ax=ax)
        ax = plt.subplot(3,1,3)
        y.plot_bse(['eps','BndsRnXs'],ax=ax)
        plt.tight_layout()
        #plt.savefig('screening_conv_plot.png')
        plt.show()
    else:
        ax = plt.subplot(3,1,1)
        y.plot_bse(['eps','BSEEhEny'],ax=ax)
        ax = plt.subplot(3,1,2)
        y.plot_bse(['eps','BSENGBlk'],ax=ax)
        ax = plt.subplot(3,1,3)
        y.plot_bse(['eps','BSENGexx'],ax=ax)
        plt.tight_layout()
        #plt.savefig('absorption_conv_plot.png')
        plt.show()
    print('done!')

if __name__ == "__main__":
    #parse options
    parser = argparse.ArgumentParser(description='Test the yambopy script.')
    parser.add_argument('-r', '--run',     action="store_true",  help='run BSE convergence calculation')
    parser.add_argument('-a', '--analyse', action="store_true",  help='analyse results data')
    parser.add_argument('-p', '--plot',    action="store_true",  help='plot the results')
    parser.add_argument('-e', '--epsilon', action="store_true",  help='converge epsilon parameters')
    parser.add_argument('-b', '--bse',     action="store_true",  help='converge bse parameters')
    parser.add_argument('-u', '--nohup',   action="store_true",  help='run the commands with nohup')
    parser.add_argument('-t', '--threads', default=1, type=int,  help='number of threads to use')

    args = parser.parse_args()
    threads = args.threads
    nohup = args.nohup

    if len(sys.argv)==1:
        parser.print_help()
        sys.exit(1)

    if args.bse:
        what = 'bse'
    else:
        what = 'screening'

    create_save()
    if args.run:     bse_convergence(what=what,threads=threads,nohup=nohup)
    if args.analyse: analyse()
    if args.plot:    plot(what)



# Copyright (c) 2018, Henrique Miranda
# All rights reserved.
#
# This file is part of the yambopy project
#
from __future__ import print_function, division
from yambopy import *
from itertools import product
from yambopy.plot import *
from yambopy.lattice import red_car

def jump_to(f,tag):
    """ Jump to a line in file
    """
    while True:
        line = f.readline()
        if tag in line:
            break
def v2str(v):
    return ("%12.8lf "*len(v))%tuple(v)

class YamboExcitonWaveFunctionXSF(object):
    """
    Class to read excitonic wavefunctions from yambo in the 3D xsf format
    """
    def __init__(self):
        self.cell = []
        self.atoms = []
        self.lattice = []
        self.initialized = False

    def read_file(self,filename):
        f = open(filename)
        jump_to(f,"PRIMVEC")
        self.lattice.append( list(map(float,f.readline().strip().split())) )
        self.lattice.append( list(map(float,f.readline().strip().split())) )
        self.lattice.append( list(map(float,f.readline().strip().split())) )

        jump_to(f,"PRIMCOORD")
        self.natoms = int(f.readline().split()[0])-1

        #read the hole position
        self.hole = list(map(float,f.readline().strip().split()))

        #read the atoms positions
        self.atoms = []
        for i in range(self.natoms):
            self.atoms.append( list(map(float,f.readline().strip().split())) )

        #get atypes
        self.atypes = np.unique([a[0] for a in self.atoms]).tolist()
        atypes_dict = dict([(a,n) for n,a in enumerate(self.atypes)])
        self.atoms = [ [atypes_dict[a[0]]]+a[1:] for a in self.atoms]

        jump_to(f,"BEGIN_DATAGRID_3D")
        self.nx, self.ny, self.nz = list(map(int, f.readline().strip().split()))
        f.readline() #ignore

        #read cell
        self.cell.append( list(map(float,f.readline().strip().split())) )
        self.cell.append( list(map(float,f.readline().strip().split())) )
        self.cell.append( list(map(float,f.readline().strip().split())) )

        #read data
        self.datagrid = np.zeros([self.nz,self.ny,self.nx])
        for k,j,i in product(list(range(self.nz)),list(range(self.ny)),list(range(self.nx))):
            self.datagrid[k,j,i] = float(f.readline())
        self.initialized = True

    def plot_slice_x(self,n):
        """ plot a slice of the 3d grid
        """
        plt.imshow(self.datagrid[:,:,n])
        plt.show()

    def plot_slice_z(self,n):
        """ plot a slice of the 3d grid
        """
        plt.imshow(self.datagrid[n,:,:])
        plt.show()

    def write_xsf(self,filename):
        f = open(filename,'w')
        #structure
        f.write('CRYSTAL\n')
        f.write('PRIMVEC\n')
        for vlat in self.lattice:
            f.write(v2str(vlat)+'\n')
        f.write('PRIMCOORD\n')
        f.write('%d 1\n'%len(self.atoms))
        f.write(v2str(self.hole)+'\n')
        for atom in self.atoms:
            f.write("%d "%self.atypes[atom[0]]+v2str(atom[1:])+'\n')
        #datagrid
        f.write('BEGIN_BLOCK_DATAGRID_3D\n')
        f.write('excitonwf\n')
        f.write('BEGIN_DATAGRID_3D\n')
        f.write('%d %d %d\n'%(self.nx,self.ny,self.nz))
        f.write('0.00 0.00 0.00\n')
        for vlat in self.lattice:
            f.write(v2str(vlat)+'\n')
        for x in self.datagrid.flatten():
            f.write('%lf\n'%x)
        f.write('END_DATAGRID_3D\n')
        f.write('END_BLOCK_DATAGRID_3D\n')
        f.close()

    def center(self):
        """ Center the atoms and excitonic wavefunction in the unit cell
        """
        self.norm = [0,0,0]
        for i in range(3):
            self.norm[i] = np.linalg.norm(self.lattice[i])

        print(self.norm)
        print(self.nx, self.ny, self.nz)

        # find the average position in each direction (geometric center)
        pos = np.zeros([3])
        for atom in self.atoms:
            pos += atom[1:]
        pos = pos/len(self.atoms)

        # center the atoms around that position using the center of the unit cell
        displecement = pos
        for atom in self.atoms:
            x,y,z = atompos + atom[1:]
            atom[1:] = np.array([x%self.nx, y%self.ny, z%self.nz])

        # center the wavefunction around that position
        new_data = np.zeros(data.shape)

    def get_data(self):
        return { "datagrid": self.datagrid.flatten().tolist(),
                 "lattice": self.lattice,
                 "atoms": self.atoms,
                 "atypes": self.atypes,
                 "hole": self.hole,
                 "nx": self.nx,
                 "ny": self.ny,
                 "nz": self.nz }

    def write_json(self):
        """ Write as a json file
        """
        JsonDumper(self.get_data(),"datagrid.json")

    def read_json_file(self,filename):
        f = open(filename,"r")
        data = json.load(f)
        f.close()

        self.read_json(data)

    def read_json(self,data):
        """ Write as a json file
        """
        self.datagrid = data["datagrid"]
        self.lattice = data["lattice"]
        self.atoms = data["atoms"]
        self.atypes = data["atypes"]
        self.nx = data["nx"]
        self.ny = data["ny"]
        self.nz = data["nz"]
        self.natoms = len(self.atoms)

        self.initialized = True

    def __str__(self):
        s = ""
        s += "lattice:\n"
        for i in range(3):
            s += ("%12.8lf "*3)%tuple(self.lattice[i])+"\n"
        s += "natoms:\n"
        s += "atoms:\n"
        for i in range(self.natoms):
            s += ("%3d "+"%12.8lf "*3)%tuple(self.atoms[i])+"\n"
        s += "atypes:\n"
        for n,a in enumerate(self.atypes):
            s += "%3d %3d\n"%(n,a)
        s += "nx: %d\n"%self.nx
        s += "ny: %d\n"%self.ny
        s += "nz: %d\n"%self.nz
        return s



# This file is part of yambopy
# Author: FP

import os
from yambopy import *
from schedulerpy import *

class CreateYamboSave():
    """
    Generation of any type of yambo SAVE folder

    Must be run outside the folder where the nscf calculation took place.

    It contains:                                                                  CASES:      
        - function to generate basic p2y+yambo SAVE ----------------------------> save_type='simple' [default]
        - functions to generate more advanced SAVEs as :

             -- fixsymm (for real-time calculations)  --------------------------> save_type='fixsymm' [also turned on by field_dir]
             -- elph    (for electron-phonon calculations) ---------------------> save_type='elph'    [also turned on by elph_path]
             -- combinations of the above, including:

                 --- expansion of elph matrix elements -------------------------> save_type='expanded_elph'
                 --- fixsymm with elph matrix elements -------------------------> save_type='fixsymm+elph'
                 --- reading of bare elph matrix elements (always when found)

    Cases supported:
        - simple, elph, expanded_elph, fixsymm, fixsymm+elph
   
    Additional relevant variables:
        - elph_path : path of dfpt electron-phonon databases
        - field_dir : direction of incoming external field

    Examples of use:
 
        :: Generate a SAVE file:
      
               CreateYamboSave(prefix,save_type='simple',nscf=nscf_dir,database=save_dir,yambo_exec_path=yambo_bin)
    
        :: Generate a SAVE file with reduced symmetries:
        
               CreateYamboSave(prefix,save_type='fixsymm',field_dir=[1,0,0])

        :: Generate a SAVE file with expanded elph matrix elements:
    
               CreateYamboSave(prefix,save_type='expanded_elph',elph_path=elph_dir)

    """
    def __init__(self,prefix,nscf='./nscf',database='./database',save_type='simple',field_dir=None,elph_path=None,MaxGvecs=None,yambo_exec_path='',printIO=True):

        list_of_possibilities = ['simple','elph','expanded_elph','fixsymm','fixsymm+elph']

        if not os.path.isdir(database): os.mkdir(database)

        #Global variables        
        self.scheduler = Scheduler.factory
        self.MaxGvecs  = MaxGvecs
        self.prefix    = prefix
        
        #Yambo executables
        if yambo_exec_path != '': yambo_exec_path+='/'
        self.yambo     = yambo_exec_path + 'yambo'
        self.yambo_rt  = yambo_exec_path + 'yambo_rt'
        self.yambo_ph  = yambo_exec_path + 'yambo_ph' 
        self.p2y       = yambo_exec_path + 'p2y'
        self.ypp       = yambo_exec_path + 'ypp'
        self.ypp_ph    = yambo_exec_path + 'ypp_ph'
        
        # Logic-important variables
        self.elph_path = elph_path
        self.field_dir = field_dir
        self.save_type = save_type
        
        #Start IO
        self.yf = YamboIO(out_name='YAMBOPY_SAVEsetup.log',out_path=database,print_to_shell=printIO)
        self.yf.IO_start()

        # [1] Logic to determine which SAVE to generate
        self.which_SAVE()
        # Up to this point we have save_type, field_dir and elph_path set correctly
        
        # [2] Generate SAVE
        if self.save_type=='simple' or self.save_type not in list_of_possibilities: 
            # [2a] Standard SAVE already initialized
            self.get_SAVE(nscf,database)
        else:
            # [2b] Standard SAVE not initialised
            self.get_SAVE(nscf,database,noinit=True)
            
            # [3] Additional operations
            if self.save_type=='elph':
                # [3a] SAVE including ndb.elph_gkkp databases
                self.get_SAVE_elph(database)
            elif self.save_type=='expanded_elph':
                # [3b] SAVE including expanded ndb.elph_gkkp_expanded databases
                self.get_SAVE_elph(database,expand=True)
            elif self.save_type=='fixsymm':
                # [3c] SAVE with symmetries removed according to an external field
                self.get_SAVE_fixsymm(database)
            elif self.save_type=='fixsymm+elph':
                # [3d] SAVE in cases [3c]+[3b]
                self.get_SAVE_fixsymm(database,with_elph=True)
            else: 
                raise ValueError('No suitable save_type found. See docstring for information on avalaible save types.')
        
        #End IO
        self.yf.IO_close()

    def which_SAVE(self):
        """
        Set logicals for execution.

        This lengthy function serves only one purpose:
        instruct the code about what to do if the user gives sloppy inputs; by default it tries to be as generous as possible.
        """

        elph_is_there = os.path.isdir('./elph_dir') or os.path.isfile('./elph_dir')

        # The user has not specified elph_path and field_dir but may have set a save_type that requires them
        if self.elph_path is None and self.field_dir is None:
            if self.save_type=='simple': 
                self.yf.msg('## Creating standard SAVE ##')
            elif self.save_type=='elph':
                if not elph_is_there: raise FileNotFoundError('elph_dir directory not specified.')
                else: self.yf.msg('## Creating SAVE with electron-phonon databases ##')
            elif self.save_type=='expanded_elph': 
                if not elph_is_there: raise FileNotFoundError('elph_dir directory not specified.')
                else: self.yf.msg('## Creating SAVE with expanded electron-phonon databases ##')
            elif self.save_type=='fixsymm':
                self.yf.msg('## Creating SAVE with fixed symmetry ##')
                self.yf.msg('Field direction not specified, setting it to [1,0,0].')
                self.field_dir = [1,0,0]
            elif self.save_type=='fixsymm+elph':
                if not elph_is_there: raise FileNotFoundError('elph_dir directory not specified.')
                else: self.yf.msg('## Creating SAVE with fixed symmetry and expanded electron-phonon databases ##')
                self.yf.msg('Field direction not specified, setting it to [1,0,0].')
                self.field_dir = [1,0,0]

        # The user has specified both elph_path and field_dir but may have set an inconsistent save_type
        elif self.elph_path is not None and self.field_dir is not None:
            if self.save_type in ['simple','elph','expanded_elph','fixsymm']:
                self.yf.msg('## Creating SAVE with fixed symmetry and expanded electron-phonon databases ##')
                self.save_type='fixsymm+elph'
            elif self.save_type=='fixsymm+elph':
                self.yf.msg('## Creating SAVE with fixed symmetry and expanded electron-phonon databases ##')

        # The user has specified only field_dir but may have set an inconsistent save_type
        elif self.elph_path is None and self.field_dir is not None:
            if self.save_type=='simple':
                self.yf.msg('## Creating SAVE with fixed symmetry ##')
                self.save_type='fixsymm'
            elif self.save_type in ['elph','expanded_elph']:
                if not elph_is_there: raise FileNotFoundError('elph_dir directory not specified.')
                else: self.yf.msg('## Creating SAVE with fixed symmetry and expanded electron-phonon databases ##')
                self.save_type='fixsymm+elph'
            elif self.save_type=='fixsymm':
                self.yf.msg('## Creating SAVE with fixed symmetry ##')
            elif self.save_type=='fixsymm+elph':
                if not elph_is_there:
                    self.yf.msg('[WARNING]: elph_dir directory not specified')
                    self.yf.msg('## Creating SAVE with fixed symmetry (but no elph databases) ##')
                    self.save_type='fixsymm'
                else: 
                    self.yf.msg('## Creating SAVE with fixed symmetry and expanded electron-phonon databases ##')

        # The user has specified only elph_path but may have set an inconsistent save_type
        elif self.elph_path is not None and self.field_dir is None:
            if self.save_type=='simple':
                self.yf.msg('## Creating SAVE with electron-phonon databases ##')
                self.save_type='elph'
            elif self.save_type=='elph':
                self.yf.msg('## Creating SAVE with electron-phonon databases ##')
            elif self.save_type=='expanded_elph':
                self.yf.msg('## Creating SAVE with expanded electron-phonon databases ##')
            elif self.save_type=='fixsymm':
                self.yf.msg('## Creating SAVE with fixed symmetry and expanded electron-phonon databases ##')
                self.yf.msg('Field direction not specified, setting it to [1,0,0].')
                self.field_dir = [1,0,0]
                self.save_type='fixsymm+elph'
            elif self.save_type=='fixsymm+elph':
                self.yf.msg('## Creating SAVE with fixed symmetry and expanded electron-phonon databases ##')
                self.yf.msg('Field direction not specified, setting it to [1,0,0].')
                self.field_dir = [1,0,0]

    def get_SAVE(self,nscf,database,noinit=False):
        """
        Generate SAVE folder from QE nscf calculation
        """
        qe_save = '%s/%s.save'%(nscf,self.prefix)
        #check if the nscf cycle is present
        if os.path.isdir(qe_save): 
            self.yf.msg('nscf calculation found!')
        else: 
            self.yf.msg('nscf calculation not found!')
            raise FileNotFoundError('nscf calculation not found!')

        #check if the SAVE folder is present
        if os.path.isdir('%s/SAVE'%database):
            self.yf.msg('SAVE database found!')
            
        #generate SAVE
        if not os.path.isdir('%s/SAVE'%database):
            self.yf.msg('preparing yambo database')
            p2y_run = self.scheduler()
            p2y_run.add_command('mkdir -p %s'%database)
            p2y_run.add_command('cd %s; %s > p2y.log ; cd -'%(qe_save,self.p2y))
            if not noinit: p2y_run.add_command('cd %s; %s > yambo.log ; cd -'%(qe_save,self.yambo))
            p2y_run.add_command('mv %s/SAVE  %s/'%(qe_save,database))
            p2y_run.run()

    def get_SAVE_elph(self,database,expand=False,filnm1='setup.in'):
        """
        Add gkkp from dfpt calculation to original SAVE folder
        """
        
        gkkp_dbs = 'ndb.elph_gkkp'
        if expand: gkkp_dbs+='_expanded'

        #check if gkkp databases are already present 
        if os.path.isfile('%s/SAVE/%s'%(database,gkkp_dbs)):
            self.yf.msg('gkkp already present.')
        else:
            #check if the elph_dir folder is present
            if not os.path.isfile('%s/elph_dir/s.dbph_000001'%self.elph_path):
                raise FileNotFoundError('problem with dbph databases at %s/elph_dir'%self.elph_path)
            else:
                self.yf.msg('Reading gkkp')
            
                filnm2 = 'gkkp.in'

                # Initialisation using yambo_ph
                y1 = YamboIn.from_runlevel('-i -V RL',executable=self.yambo_ph,filename=filnm1,folder=database)
                y1.arguments.append('BSEscatt')
                if self.MaxGvecs is not None: y1['MaxGvecs'] = self.MaxGvecs
                y1.write('%s/%s'%(database,filnm1))
                yamboph_run = self.scheduler()
                if not os.path.islink('%s/elph_dir'%database): yamboph_run.add_command('cd %s ; ln -s %s/elph_dir . ; cd -'%(database,self.elph_path))
                yamboph_run.add_command('cd %s ; %s -F %s -J ./elph_dir ; cd -'%(database,self.yambo_ph,filnm1))
                yamboph_run.run()

                # Reading and expansion of gkkp
                yph = YamboIn.from_runlevel('-gkkp',executable=self.ypp_ph,filename=filnm2,folder=database)
                if expand:
                    yph.arguments.append('GkkpExpand')
                    self.yf.msg('    expanding gkkp in the full BZ')
                yph['DBsPATH'] = "./elph_dir"
                
                # Check if bare gkkp are present
                if os.path.isfile('%s/elph_dir/s.dbph_bare_000001'%self.elph_path):
                    self.yf.msg('    reading also bare gkkp')
                    yph.arguments.append('GkkpReadBare')
                yph.write('%s/%s'%(database,filnm2))
                
                yppph_run = self.scheduler()
                yppph_run.add_command('cd %s ; %s -F %s; cd -'%(database,self.ypp_ph,filnm2))
                yppph_run.run()
                    
                if not os.path.isfile('%s/SAVE/%s'%(database,gkkp_dbs)):
                    self.yf.msg('[ERROR] %s databases not created. Check the logs.'%gkkp_dbs)
                    raise FileNotFoundError('%s databases not created. Check the logs.'%gkkp_dbs)

    def get_SAVE_fixsymm(self,database,with_elph=False):
        """
        Generate SAVE folder with reduced symmetries starting from original SAVE
        """
        filnm1 = 'setup.in'
        filnm2 = 'fixsymm.in'
        #check if symmetries have been removed
        if os.path.isdir('%s/FixSymm'%database):
            self.yf.msg('FixSymm folder found!')
        if not os.path.isdir('%s/FixSymm'%database):
            
            if not with_elph:
                # Initialisation using yambo_rt
                y1 = YamboIn.from_runlevel('-i -V RL',executable=self.yambo_rt,filename=filnm1,folder=database)
                if self.MaxGvecs is not None: y1['MaxGvecs'] = self.MaxGvecs
                y1.write('%s/%s'%(database,filnm1))
                yambort_run = self.scheduler()
                yambort_run.add_command('cd %s ; %s -F %s; cd -'%(database,self.yambo_rt,filnm1))
                yambort_run.run()
            else:
                # Get expanded gkkp
                self.get_SAVE_elph(database,expand=True,filnm1=filnm1)

            # Removing symmetries
            self.yf.msg('Removing symmetries')
            y2 = YamboIn.from_runlevel('-y',executable=self.ypp,filename=filnm2,folder=database)
            y2['Efield1']=self.field_dir
            y2.arguments.append('RmTimeRev')
            y2.write('%s/%s'%(database,filnm2))
            ypp_run = self.scheduler()
            ypp_run.add_command('cd %s ; %s -F %s ; cd -'%(database,self.ypp,filnm2))
            
            # Re-initialisation
            if not with_elph:
                ypp_run.add_command('cd %s/FixSymm ; %s ; cd -'%(database,self.yambo_rt))
            else:
                ypp_run.add_command('cd %s/FixSymm/SAVE ; mv ../../SAVE/ndb.elph_gkkp_expanded* . ; cd -'%database)
                ypp_run.add_command('cd %s/FixSymm ; cp ../%s . ; cd -'%(database,filnm1))
                ypp_run.add_command('cd %s/FixSymm ; %s -F %s -J SAVE ; cd -'%(database,self.yambo_ph,filnm1))
            ypp_run.run()



# Copyright (c) 2018, Henrique Miranda
# All rights reserved.
#
# This file is part of the yambopy project
#
import numpy as np
from yambopy.tools.string import marquee
from yambopy.plot.plotting import add_fig_kwargs
from qepy.lattice import Path

def exagerate_differences(ks_ebandsc,ks_ebandsp,ks_ebandsm,d=0.01,exagerate=5):
    """
    Take three different band-structures with some parameter changing
    and return an exagerated version for plotting.
    This is used for finite differences w.r.t. atomic positions for example
    """
    #calculate central finite difference
    fdc_ks = (ks_ebandsp-ks_ebandsm)/(d*2)
    fdp_ks = (ks_ebandsp-ks_ebandsc)/d
    fdm_ks = (ks_ebandsm-ks_ebandsc)/d

    #exagerate differences
    ks_ebandsp = ks_ebandsc + fdp_ks*d*exagerate
    ks_ebandsm = ks_ebandsc + fdm_ks*d*exagerate

    #set colors again
    ks_ebandsp.set_kwargs(c='brown')
    ks_ebandsm.set_kwargs(c='tomato')

    return ks_ebandsc,ks_ebandsp,ks_ebandsm

def apply_scissor_shift(eigenvalues,scissor,n_val):
    """
    Apply scissor shift to band structure
    
    Input:
      :: eigenvalues -> np array of dimensions (Nk,Nb) or (Ns,Nk,Nb)
      :: scissor -> [shift, stretch_cond, stretch_val]
      :: n_val -> number of valence bands 

      NB: make sure eigenvalues and scissor have same units!        

    Returns shifted eigenvalues
    """
    from copy import deepcopy

    # Dimensionality including spin
    # In this case reshape/concatenate to get (Nkponts,Nbands) array
    if len(eigenvalues.shape)==3:
        Nspin, Nkpoints, Nbands = eigenvalues.shape
        if Nspin>1: raise NotImplementedError("Scissor for spin-polarised bands not yet implemented.")
        else: eigen_to_shift = deepcopy(eigenvalues[0])
     # If original dimensionality is (Nkpoints,Nbands) work with original array
    else: 
        Nkpoints, Nbands = eigenvalues.shape
        eigen_to_shift = deepcopy(eigenvalues)

    # Actual scissor operator code          
    aux = np.zeros((Nkpoints,Nbands))
    top_v, bottom_c = eigen_to_shift[:,n_val-1], eigen_to_shift[:,n_val]
    ind_k_dir_gap = np.argmin(bottom_c-top_v)
    ev_max, ec_min = top_v[ind_k_dir_gap], bottom_c[ind_k_dir_gap]

    for ib in range( Nbands ):
        if ib<n_val: aux[:,ib] = ev_max-(ev_max-eigen_to_shift[:,ib])*scissor[2]
        else:        aux[:,ib] = ec_min+scissor[0]+(eigen_to_shift[:,ib]-ec_min)*scissor[1]

    if len(eigenvalues.shape)==3: final_eigen = np.expand_dims(aux, axis=0)
    else:                         final_eigen = aux

    return final_eigen

class YambopyBandStructure():
    """
    Class to plot bandstructures
    I include spin projection (to be improve and checked)  AMS
    """
    _colormap = 'rainbow'

    def __init__(self,bands,kpoints,kpath=None,fermie=0,weights=None,spin_proj=None,**kwargs):
        self.bands = np.array(bands)
        self.weights = np.array(weights)     if weights   is not None else None
        self.spin_proj = np.array(spin_proj) if spin_proj is not None else None 
        self.kpoints = np.array(kpoints)
        self.kwargs = kwargs
        self.kpath = kpath
        self.fermie = fermie
        self._xlim = None
        self._ylim = None

    @property
    def nbands(self):
        nkpoints, nbands = self.bands.shape
        return nbands

    @property
    def nkpoints(self):
        nkpoints, nbands = self.bands.shape
        return nkpoints

    @property
    def xlim(self):
        if self._xlim is None: return (min(self.distances),max(self.distances))
        return self._xlim

    @property
    def ylim(self):
        if self._ylim is None: return (np.min(self.bands)-self.fermie,np.max(self.bands)-self.fermie)
        return self._ylim

    @classmethod
    def from_dict(cls,d):
        path = Path.from_dict(d['kpath'])
        instance = cls(d['bands'],d['kpoints'],kpath=path,
                       fermie=d['fermie'],weights=d['weights'],**d['kwargs'])
        instance._xlim = d['_xlim']
        instance._ylim = d['_ylim']
        return instance

    @classmethod
    def from_json(cls,filename):
        import json
        with open(filename,'r') as f:
            d = json.load(f)
        return cls.from_dict(d)

    def as_dict(self):
        """ Return the data of this object as a dictionary
        """
        d = { 'bands': self.bands.tolist(),
              'weights': self.weights.tolist() if self.weights is not None else None,
              'kpoints': self.kpoints.tolist(),
              'kwargs': self.kwargs,
              'kpath': self.kpath.as_dict() if self.kpath is not None else None,
              'fermie': self.fermie,
              '_xlim': self._xlim,
              '_ylim': self._ylim }
        return d 

    def write_json(self,filename):
        """serialize this class as a json file"""
        import json
        with open(filename,'w') as f:
            json.dump(self.as_dict(),f)

    def set_fermi(self,valence):
        """simple function to set the fermi energy given the number of valence bands
        """
        self.fermie = np.max(self.bands[:,valence-1])
        self.set_ylim(None)

    def set_energy_offset(self,energy_offset):
        """simple function to rigid-shift the bands
        """
        self.fermie = energy_offset
        self.set_ylim(None)

    def set_xlim(self,xlim):
        self._xlim = xlim

    def set_ylim(self,ylim):
        self._ylim = ylim

    def set_ax_lim(self,ax,fermie=0,ylim=None,xlim=None):
        if xlim is None: xlim = self.xlim
        if ylim is None: ylim = self.ylim
        ax.set_xlim(xlim[0],xlim[1])
        ax.set_ylim(ylim[0]-self.fermie,ylim[1]-self.fermie)

    @property
    def distances(self):
        if not hasattr(self,"_distances"):
            self._distances = [0]
            distance = 0
            for nk in range(1,len(self.kpoints)):
                distance += np.linalg.norm(self.kpoints[nk]-self.kpoints[nk-1])
                self._distances.append(distance)
        return self._distances

    def as_list(self,bands=None):
        yl = YambopyBandStructureList([self])
        if bands: yl.append(bands)
        return yl

    @add_fig_kwargs    
    def plot(self,title=None):
        """return a matplotlib figure with the plot"""
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(1,1,1)
        self.plot_ax(ax)
        if title: ax.title(title)
        return fig

    def set_kwargs(self,**kwargs):
        self.kwargs.update(kwargs)

    def get_kwargs(self,**kwargs):
        from copy import deepcopy
        fullkwargs = deepcopy(self.kwargs)
        fullkwargs.update(kwargs)
        return fullkwargs

    def add_kpath_labels(self,ax):
        """
        Add vertical lines at the positions of the high-symmetry k-points
        I did a modification that I don't like much. To be corrected
        """
        if self.kpath is None:
            ax.xaxis.set_ticks([])
            return 
        for kpoint, klabel, distance in self.kpath:
            ax.axvline(distance,c='k',ls='--',lw=0.5)
        ax.axvline(0.0,c='k',ls='-',lw=0.0)
        ax.axvline(distance,c='k',ls='-',lw=0.0)
        self.kpath.set_xticks(ax)

    def plot_ax(self,ax,xlim=None,ylim=None,size=1.,ylabel=r'$\epsilon_{n\mathbf{k}}$ [eV]', alpha_weights=0.5,legend=False,**kwargs):
        """Receive an intance of matplotlib axes and add the plot"""
        import matplotlib.pyplot as plt
        kwargs = self.get_kwargs(**kwargs)
        fermie = kwargs.pop('fermie',self.fermie)

        # Set kwargs
        c_bands   = kwargs.pop('c_bands',None)
        c_weights = kwargs.pop('c_weights',None)
        label   = kwargs.pop('label',None)
        lw_label  = kwargs.pop('lw_label',None)
        marker = kwargs.pop('marker',None)
        linestyle = kwargs.pop('linestyle',None)

        # I choose a colormap for spin
        color_map  = plt.get_cmap('seismic')

        x = self.distances
        y = self.bands.T-fermie
        for ib,band in enumerate(self.bands.T):
            x = self.distances
            y = band-fermie
            ax.plot(x,y,c=c_bands,lw=lw_label,marker=marker,linestyle=linestyle,label=label if ib == 0 else "_nolegend_")
            # fill between 
            if self.weights is not None:

                dy = self.weights[:,ib]*size
                ax.fill_between(x,y+dy,y-dy,alpha=alpha_weights,color=c_weights,linewidth=0,label=label)

        self.set_ax_lim(ax,fermie=fermie,xlim=xlim,ylim=ylim)
        ax.set_ylabel(ylabel)
        self.add_kpath_labels(ax)
        if legend: ax.legend()

    def plot_spin_ax(self,ax,xlim=None,ylim=None,ylabel=r'$\epsilon_{n\mathbf{k}}$[eV]',alpha_weights=0.5,spin_proj_bands=None,legend=False,**kwargs):
        """Receive an intance of matplotlib axes and add the plot"""
        #
        # There is a problem with the number of points in the k-poitns path
        import matplotlib.pyplot as plt
        kwargs = self.get_kwargs(**kwargs)
        fermie = kwargs.pop('fermie',self.fermie)

        # Set kwargs
        c_bands   = kwargs.pop('c_bands',None)
        c_weights = kwargs.pop('c_weights',None)
        label   = kwargs.pop('label',None)
        lw_label  = kwargs.pop('lw_label',None)
        marker = kwargs.pop('marker',None)
        linestyle = kwargs.pop('linestyle',None)
        n_valence = kwargs.pop('n_valence',None)
        weight_option = kwargs.pop('weight_option',None)

        # I choose a colormap for spin
        color_map  = plt.get_cmap('PiYG')

        x = self.distances
        y = self.bands.T-fermie
        nk_distance = len(self.distances)
        for ib,band in enumerate(self.bands.T):
            x = self.distances
            y = band-fermie
            ax.plot(x,y,c=c_bands,lw=lw_label,marker=marker,linestyle=linestyle,label=label if ib == 0 else "_nolegend_")

            # dot
            if self.weights is not None:
               dy = self.weights[:,ib]*size*1000
               color_spin = spin_proj_bands[:,n_valence + ib] + 0.5 # I renormalize 0 => down; 1 => up
               ax.scatter(x,y,s=abs(dy),c=color_spin,cmap=color_map,edgecolors='none',zorder=2,rasterized=True) 

        self.set_ax_lim(ax,fermie=fermie,xlim=xlim,ylim=ylim)
        ax.set_ylabel(ylabel)
        self.add_kpath_labels(ax)
        if legend: ax.legend()

    def __add__(self,y):
        """Add the bands of two systems together"""
        #add some consistency check
        bands = self.bands + y.bands
        return YambopyBandStructure(bands,self.kpoints,kpath=self.kpath,fermie=self.fermie+y.fermie,**self.kwargs)
 
    def __sub__(self,y):
        """Subtract the bands of two systems together"""
        #add some consistency check
        bands = self.bands - y.bands
        return YambopyBandStructure(bands,self.kpoints,kpath=self.kpath,fermie=self.fermie-y.fermie,**self.kwargs)

    def __mul__(self,y):
        """Scale the bands of the system"""
        #add some consistency check
        bands = self.bands*y
        return YambopyBandStructure(bands,self.kpoints,kpath=self.kpath,fermie=self.fermie*y,**self.kwargs)

    def __truediv__(self,y):
        """Scale the bands of the system"""
        #add some consistency check
        bands = self.bands/y
        return YambopyBandStructure(bands,self.kpoints,kpath=self.kpath,fermie=self.fermie/y,**self.kwargs)
    
    def __str__(self):
        lines = []; app = lines.append
        app('nkpoints: %d'%self.nkpoints)
        app('nbands: %d'%self.nbands)
        app('has kpath: %s'%hasattr(self,'kpath'))
        return "\n".join(lines)

class YambopyBandStructureList():
    """This class contains a list of band-structure classes and is responsible for plotting them together"""
    def __init__(self,bandstructures):
        self.bandstructures = bandstructures

    @classmethod
    def from_pickle(cls,filename):
        import pickle
        with open(filename,'rb') as f:
            ybs = pickle.load(f)
        return ybs

    @property
    def has_legend(self):
        return any(['label' in bandstructure.kwargs for bandstructure in self.bandstructures])

    @property
    def nbandstructures(self):
        return len(self.bandstructures)

    @property
    def xlim(self):
        low_xlim = [bandstructure.xlim[0] for bandstructure in self.bandstructures]
        top_xlim = [bandstructure.xlim[1] for bandstructure in self.bandstructures]
        return (np.min(low_xlim),np.max(top_xlim))

    @property
    def ylim(self):
        low_ylim = [bandstructure.ylim[0] for bandstructure in self.bandstructures]
        top_ylim = [bandstructure.ylim[1] for bandstructure in self.bandstructures]
        return (np.min(low_ylim),np.max(top_ylim))

    def as_dict(self):
        bandstructures_list=[]
        for bandstructure in self.bandstructures:
            bandstructures_list.append(bandstructure.as_dict())
        return bandstructures_list

    @classmethod
    def from_json(cls,filename):
        import json
        with open(filename,'r') as f:
            d = json.load(f)
        return cls.from_dict(d)

    @classmethod
    def from_dict(cls,d):
        bandstructures = []
        for bandstructure_dict in d:
            bandstructure = YambopyBandStructure.from_dict(bandstructure_dict)
            bandstructures.append(bandstructure)
        return cls(bandstructures)

    def __getitem__(self,idx):
        return self.bandstructures[idx]
    
    def append(self,bands):
        if isinstance(bands,list): self.bandstructure.extend(bands)
        self.bandstructures.append(bands)

    def add_bandstructure(self,bandstructures,**kwargs):
        """
        Add a bandstructure to bandstructure set
        """
        if not isinstance(bandstructures,list): bandstructures = [bandstructures]

        #add arguments in all the structures
        for bandstructure in bandstructures:
            bandstructure.kwargs.update(kwargs)

        #extend current bandstructure set
        self.bandstructures.extend(bandstructures)

    def plot_ax(self,ax,legend=True,xlim=None,ylim=None,**kwargs):
        title = kwargs.pop('title',None)
        for i,bandstructure in enumerate(self.bandstructures):
            bandstructure.plot_ax(ax,**kwargs)
        if xlim is None: ax.set_xlim(self.xlim)
        else:            ax.set_xlim(xlim)
        if ylim is None: ax.set_ylim(self.ylim)
        else:            ax.set_ylim(ylim)
        if title: ax.set_title(title)
        if legend and self.has_legend: ax.legend()

    def set_fermi(self,valence):
        """Find the Fermi energy of all the bandstructures by shifting the Fermi energy"""
        for bandstructure in self.bandstructures:
            bandstructure.set_fermi(valence)

    @add_fig_kwargs
    def plot(self,figsize=None,**kwargs):
        import matplotlib.pyplot as plt
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(1,1,1)
        self.plot_ax(ax,**kwargs)
        return fig
    
    def get_color(self,i,colormap='gist_rainbow'):
        colormap = self.get_colormap(colormap=colormap) 
        return colormap[i]

    def get_colormap(self,colormap='gist_rainbow'):
        """get a list of colors for each plot"""
        import matplotlib.pyplot as plt
        cmap = plt.get_cmap(colormap) #get color map
        return [cmap(i) for i in np.linspace(0, 1, self.nbandstructures)]

    def write_json(self,filename):
        """serialize this class as a json file"""
        import json
        with open(filename,'w') as f:
            json.dump(self.as_dict(),f)

    def pickle(self,filename):
        import pickle
        with open(filename,'wb') as f:
            pickle.dump(self,f)

    def __str__(self):
        lines = []; app = lines.append
        app(marquee(self.__class__.__name__))
        app('nbandstructures: %d'%self.nbandstructures)
        return "\n".join(lines)



#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: FP
#
# This file is part of the yambopy project
#
import numpy as np

def get_spectra(energies,constant=1.,weights=None,residuals=None,broadening=0.01,emin=0,emax=10,estep=0.01):
    """
    DOS and spectral functions.

    This class computes the generic quantity: 

            D(w)= C sum\_i^N_i wa\_i |D\_i|^2 delta_b(w - dE\_i)
    
    Inputs:
    * dE_i:    pole energies in eV, must be an array of shape (N_momenta, N_states) or (N_states)
    * C:       dimensional constant [default: 1.]
    * wa_i:    weights of the i-th contribution (e.g. because of sums over k or q points) [default: 1./N_momenta]
    * |D_i|^2: matrix elements squared [default: 1.], must be an array of shape (N_momenta, N_states) or (N_states)
    * b:       Lorentzian broadening of the delta function [default: 0.01 eV, note that b is the HWHM, i.e. 0.5*FWHM]
    * emin, emax, estep:  energy range parameters in eV [default: (0,10, 0.01)]
    
    Outputs:
    * w:       energy range in eV
    * D(w):    DOS or spectral function

    Possible applications:
    * DOS:                 D(w)    =   sum\_vk  w_\k              delta_b(w - e\_vk)
    * JDOS:                D(w)    =   sum\_cvk w_\k              delta_b(w - (e_\ck-e\_vk))
    * IP abs.:             D(w)    = C sum\_cvk w_\k |D_\cvk|^2   delta_b(w - (e_\ck-e\_vk))
    * BSE abs.:            D(w)    = C sum\_l        |R_\l|^2     delta_b(w - E_\l)
    * ph. DOS:             D(w)    =   sum\_mq  w_\q              delta_b(w - w\_mq)
    * ph. Eliashberg fct.: D_ik(w) =   sum\_mq  w_\q T^\ik_\mq(T) delta_b(w - w\_mq)  [ T^\ik_\mq(T)=(\sum_j|g^\ijk^\mq|^2)(2N_\mq(T)+1), only FM term]
    """
    # Energy range
    w = np.arange(emin,emax,estep)

    # Check if we are integrating over momenta or not
    try:
        nmomenta = energies.shape[0]
        nstates  = energies.shape[1]
    except IndexError:
        nmomenta = None
        nstates = energies.shape[0]

    # This is always the case for single-particle quantities
    if nmomenta is not None: 

        if weights is None:   weights   = np.ones(nmomenta)/nmomenta
        if residuals is None: residuals = np.ones((nmomenta,nstates))

        # This is a (nmomenta, nstates, len(w)) array
        energy_poles = w[None, None, :] - energies[:, :, None]

        # Compute the DOS / spectral function
        D = constant*np.einsum('k,kn,knw->w', weights, residuals, broadening / (energy_poles**2 + broadening**2), optimize=True)  

    # Case of the q=0 BSE absorption spectrum 
    else:

        if weights is None:   weights   = np.ones(nstates)
        if residuals is None: residuals = np.ones(nstates)

        # This is a (nstates, len(w)) array
        energy_poles = w[None, :] - energies[:, None]

        # Compute the DOS / spectral function
        D = constant*np.einsum('n,n,nw->w', weights, residuals, broadening / (energy_poles**2 + broadening**2),optimize=True)

    return w,D



import os
import numpy as np
from netCDF4 import *
from glob import glob
import argparse

"""
Script to calculate off-diago corrections of yambo ndb.QP databases in order to plot band structure.
For reference, check Appendix A of PHYSICAL REVIEW RESEARCH 2, 043105 (2020)

Inputs:
 1. --fld_diag='path/to/folder/with/diago/dbs' [e.g., the diag/]
 2. --fld_offdiag='path/to/folder/with/offdiagoold/dbs' [e.g., the offdiag/ndb.QP]

This script will prompt the user to go through with updating the dbs.
"""

def get_db(fldr):
    """
    Read NetCDF4 database
    """
    # FP: Not that this function could be replaced by YamboQPDB calls as:
    # $> from yambopy import *
    # $> ds = YamboQPDB.from_db(folder='fldr')
    try: ds=Dataset(f'{fldr}/ndb.QP','r+', format = 'NETCDF4')
    except: raise FileNotFoundError(f'{fldr} Database not found')
    table = np.asarray(ds['QP_table']).T  # [n m k]
    kpts = np.asarray(ds['QP_kpts']).T    # [kx, ky, kz]
    E = np.asarray(ds['QP_E']) # [Re(<n|Sigma|m>) Im(<n|Sigma|m>)]
    Eo = np.asarray(ds['QP_Eo']) # [E_DFT]
    Z = np.asarray(ds['QP_Z'])
    return ds,table, kpts, E, Eo,Z

def create_newdb(flddiag,fldoffdiag):
    """
    Print info for the user and ask if they want to go through with the change
    """
    # Size of the subspace to diagonalize
    dsdiag, table_large, kpts_large, E_large, Eo_large, Z_large = get_db(flddiag)
    print("diago database found:")
    print("=====================")

    dsoffdiag, table_small, kpts_small, E_small, Eo_small, Z_small = get_db(fldoffdiag)
    print("offdiago database found:")
    print("=====================")
    
    N = int(max(table_small[:,0]) - min(table_small[:,0]) + 1)
    # First state
    E_small_diag = np.zeros_like(E_small)
    E_large_new = E_large;
    for ik in range(1,len(kpts_small)+1): 
        cond = np.where(np.equal(table_small[:,2],ik))
        E_mat = np.reshape(E_small[:,0][cond],(N,N),order='F') + 1j*np.reshape(E_small[:,1][cond],(N,N),order='F')
        Eo_mat = np.reshape(Eo_small[cond],(N,N),order='F')
        #   Remove Z
        Z_mat = np.reshape(Z_small[:,0][cond],(N,N),order='F')
        E_mat = Eo_mat + (E_mat - Eo_mat)/Z_mat    
        E_eig_diag,dumb = np.linalg.eig(E_mat)
        E_eig = np.diag(E_eig_diag)
        Emat_order = np.argsort(np.real(np.diag(E_mat)))
        Eeig_order = np.argsort(np.real(np.diag(E_eig)))
        E_mat_diag_Re=np.real(E_eig_diag[Eeig_order])
        E_mat_diag_Im=np.imag(E_eig_diag[Eeig_order])
        #   Recompose the diagonal matrix, preserving the order of the original
        #   eigenstates, we assume that the corrections do not induce a
        #   crossover.
        dumb = np.copy(Emat_order)
        Emat_order2 = np.argsort(dumb)
        E_eig = np.diag(E_mat_diag_Re[Emat_order2])+1j*np.diag(E_mat_diag_Im[Emat_order2])
        #   % Put the Z correction back
        E_eig = Eo_mat + Z_mat*(E_eig - Eo_mat)
        #print(np.concatenate(np.reshape(np.real(E_eig),N*N)))
        #print(np.reshape(np.imag(E_eig),(N*N,1)).flatten())
        E_small_diag[:,0][cond] = np.reshape(np.real(E_eig),(N*N,1),order='F').flatten()
        E_small_diag[:,1][cond] = np.reshape(np.imag(E_eig),(N*N,1),order='F').flatten()
        #E_small_diag=E_small_diag.T
        for n in range(int(min(table_small[:,0])),int(max(table_small[:,0])+1)):
            cond_large2 = np.where(np.equal(table_large[:,2],ik),1,0)
            cond_large0 = np.where(np.equal(table_large[:,0],n),1,0)
            cond_small0 = np.where(np.equal(table_small[:,0],n),1,0)
            cond_small1 = np.where(np.equal(table_small[:,1],n),1,0)
            cond_small2 = np.where(np.equal(table_small[:,2],ik),1,0)
            index_large_new = np.where(cond_large2&cond_large0)[0].tolist()
            index_small_diag = np.where(cond_small2 & cond_small1 & cond_small0)[0].tolist()
            E_large_new[index_large_new] = E_small_diag[index_small_diag]

    filename_out = 'ndb.QP'
    os.system(f'cp {flddiag}/ndb.QP {filename_out}')
    dsout=Dataset(filename_out,'a', format = 'NETCDF4')
    print (f'Writing in {filename_out} \n')
    print (E_large_new)
    var = dsout['QP_E']                     # [Re(<n|Sigma|m>) Im(<n|Sigma|m>)]
    var =E_large_new 
    dsout.close()
    dsoffdiag.close()
    dsdiag.close()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Create a new diago dbs after a GW subspace calculation numbers in yambo databases')
    parser.add_argument('-d','--fld_diag', type=str, default="./diago", help='<Required> Path to folder with the ndb.QP diago-database   (Default is ./diago)',required=True)
    parser.add_argument('-o','--fld_offdiag', type=str,default ="./offdiago",help='<Required> Path to folder with the ndb.QP offdiagodiago-database   (Default is ./diago)',required=True)
    args = parser.parse_args()

    fld_diag = args.fld_diag
    fld_offdiag = args.fld_offiag

    create_newdb(fld_diag,fld_offdiag)



# Copyright (c) 2018, Henrique Miranda
# All rights reserved.
#
# This file is part of the yambopy project
#
from yambopy import *
import shutil
ha2ev  = 27.211396132

class YamboGreenDB(object):
    """
    Read the green's functions calculated using yambo
    These green's functions describe the spectral function of the quasiparticles.
    The quasi-particles can be from electron-phonon or GW calculations
    """
    def __init__(self,filename='ndb.G',folder='.'):
        self.folder = folder
        self.filename = "%s/%s"%(folder,filename)

        #read em1s database
        try:
            database = Dataset(self.filename, 'r')
        except:
            raise IOError("Error opening %s in YamboGreenDB"%self.filename)

        #read the Green's functions energies
        re,im = database.variables['Green_Functions_Energies'][:]
        self.energies = (re+im*1j).T

        #read the Green's Functions
        re,im = database.variables['Green_Functions'][:]
        self.green = (re+im*1j).T
        
        #read the self-energy operator
        re,im = database.variables['SE_Operator'][:]
        self.se = (re+im*1j).T

        self.nqps, self.nenergies = self.green.shape

        #read QP_table
        qptable = database.variables['QP_table'][:].astype(int)
        self.band1, self.band2, self.kindex = qptable
        self.bandmax = max(self.band1)
        self.bandmin = min(self.band1)

        #qp dictionary
        self.qp_dict = {}
        for nqp,(b1,b2,kindex) in enumerate(qptable.T):
            self.qp_dict[(b1,b2,kindex)] = nqp

        #read QP_kpts
        kpts = database.variables['QP_kpts'][:].T
        self.qpoints = kpts.shape

    def plot(self,ax,kpt=0,band=0,what='SE',e0=None,**kwargs):
        """
        Plot quantities from this database
        """
        nqp = self.qp_dict[(band,band,kpt)]
        
        x = self.energies[nqp]
        options = {'SE':self.se,
                   'green':self.green}
        y = options[what][nqp]

        #get band and k-point
        band = self.band1[nqp]
        kpt  = self.kindex[nqp]

        ax.set_title('kpt=%d band=%d'%(kpt,band))
        ax.plot(x.real,y.real,label='Re(%s)'%what,**kwargs)
        ax.plot(x.real,y.imag,label='Im(%s)'%what,**kwargs)
        if e0 is not None:
            ax.plot(x.real,e0[nqp]-x.real)

            #plot 0
            ax.axhline(0,c='k',lw=1)

            #set axis
            rmin, rmax = min(y.real),max(y.real)
            imin, imax = min(y.imag),max(y.imag)
            ax.set_ylim(min(rmin,imin),max(rmax,imax))

    def modQP(self,filename_reference,filename_new):
        """
        Take a QP file as reference and modify the values of the energies, lifetimes and Z factors
        according to the ones calculated from ndb.Green.
        
        Arguments:
            filename_reference : name of the reference file
            filename_new : name of the new file with the calculated data
        """

        #copy ref file to new file
        shutil.copy(filename_reference, filename_new)

        #read QP file
        qp = Dataset(filename_new,'r+')

        #check dimensions
        #print qp.variables['QP_E_Eo_Z'][:].shape
        #print self.eqp.shape
        #print self.z.shape

        qp.variables['QP_E_Eo_Z'][0,:,0] = self.eqp.real
        qp.variables['QP_E_Eo_Z'][1,:,0] = self.eqp.imag
        qp.variables['QP_E_Eo_Z'][0,:,2] = self.z.real
        qp.variables['QP_E_Eo_Z'][1,:,2] = self.z.imag

        #write 
        qp.close()

    def getQP(self,e0,bandmin=None,bandmax=None,debug=False,secant=True,braket=None):
        """
        Get quasiparticle states
    
        Arguments:
        e0 -> bare eigenvalues in eV
        """
        from scipy.optimize import bisect, newton
        from scipy.interpolate import interp1d
        from scipy.misc import derivative

        #check if the eigenvalues have the correct dimensions
        if len(e0) != self.nqps:
            raise ValueError('Wrong dimensions in bare eigenvalues')

        #in case something is strange we plot the stuff
        def error(nqp):
            ax = plt.gca()

            #plot 0
            ax.axhline(0,c='k',lw=1)
            
            #se limits
            semin = min(self.se[nqp].real)
            semax = max(self.se[nqp].real)
            plt.ylim(semin,semax)

            #plot self energy
            self.plot(ax,nqp=nqp)

            #plot omega-e0
            emin = min(self.energies[nqp].real)
            emax = max(self.energies[nqp].real)
            x = np.linspace(emin,emax,100)
            plt.plot(x,x-e0[nqp])

            #plot imaginary part of greens funciton
            x = self.energies[nqp].real
            y = self.green[nqp].imag
            plt.plot(x,y/max(y)*semax)
    
            #plot eqp
            #plt.axvline(self.eqp[nqp],lw=1)
            #plt.axvline(e0[nqp],lw=1)

            plt.legend(frameon=False)
            plt.show()

        if bandmin is None: bandmin = self.bandmin
        if bandmax is None: bandmax = self.bandmax
 
        self.eqp = np.zeros([self.nqps],dtype=complex) 
        self.z   = np.zeros([self.nqps],dtype=complex) 
        for nqp in range(self.nqps):

            band = self.band1[nqp]
            kpt  = self.kindex[nqp]
            if debug: print("%3d %3d %3d %8.4lf"%(nqp, kpt, band, e0[nqp]))

            if not (bandmin <= band <= bandmax):
                continue

            #get x and y
            x = self.energies[nqp].real
            y = self.se[nqp]

            #interpolate real part of function
            f = interp1d(x,y.real-x+e0[nqp],kind='slinear')

            #find zero
            if secant:
                try:
                    eqp = newton(f,e0[nqp],maxiter=200)
                except ValueError as msg:
                    print(msg)
                    if debug: error(nqp)
            else:
                if braket:
                    emin = e0[nqp]-braket
                    emax = e0[nqp]+braket
                else:
                    emin = min(x)
                    emax = max(x)

                eqp = bisect(f,emin,emax) 

            #interpolate whole function
            f = interp1d(x,y)

            #calculate Z factors
            dse = derivative(f,eqp,dx=1e-8)
            z = 1./(1-dse)

            #find Im(Se(EQP)) which corresponds to the lifetime
            lif = f(eqp).imag
            eqp += 1j*lif

            #store values
            self.eqp[nqp] = eqp
            self.z[nqp] = z           

            #cehck for potential errors
            if z>1 and debug:
                print(z)
                error(nqp)            

        return self.eqp, self.z

    def __str__(self):
        s = ""
        s += "nenergies: %d\n"%self.nenergies
        s += "nqps:      %d\n"%self.nqps
        s += "bandmin:   %d\n"%self.bandmin
        s += "bandmax:   %d"%self.bandmax
        return s

